import os
import re
import typing
import shutil
from . import utils, sh
from . import bootstrap
from . import log
from . import constants
from . import corosync
from . import xmlutil
from . import watchdog
from . import parallax
from .service_manager import ServiceManager
from .sh import ShellUtils

logger = log.setup_logger(__name__)


class SBDUtils:
    '''
    Consolidate sbd related utility methods
    '''
    @staticmethod
    def get_sbd_device_metadata(dev, timeout_only=False, remote=None) -> dict:
        '''
        Extract metadata from sbd device header
        '''
        sbd_info = {}
        try:
            out = sh.cluster_shell().get_stdout_or_raise_error(f"sbd -d {dev} dump", remote)
        except Exception:
            return sbd_info

        pattern = r"UUID\s+:\s+(\S+)|Timeout\s+\((\w+)\)\s+:\s+(\d+)"
        matches = re.findall(pattern, out)
        for uuid, timeout_type, timeout_value in matches:
            if uuid and not timeout_only:
                sbd_info["uuid"] = uuid
            elif timeout_type and timeout_value:
                sbd_info[timeout_type] = int(timeout_value)
        return sbd_info

    @staticmethod
    def get_device_uuid(dev, node=None):
        '''
        Get UUID for specific device and node
        '''
        res = SBDUtils.get_sbd_device_metadata(dev, remote=node).get("uuid")
        if not res:
            raise ValueError(f"Cannot find sbd device UUID for {dev}")
        return res

    @staticmethod
    def compare_device_uuid(dev, node_list):
        '''
        Compare local sbd device UUID with other node's sbd device UUID
        '''
        if not node_list:
            return
        local_uuid = SBDUtils.get_device_uuid(dev)
        for node in node_list:
            remote_uuid = SBDUtils.get_device_uuid(dev, node)
            if local_uuid != remote_uuid:
                raise ValueError(f"Device {dev} doesn't have the same UUID with {node}")

    @staticmethod
    def verify_sbd_device(dev_list, compare_node_list=[]):
        if len(dev_list) > SBDManager.SBD_DEVICE_MAX:
            raise ValueError(f"Maximum number of SBD device is {SBDManager.SBD_DEVICE_MAX}")
        for dev in dev_list:
            if not utils.is_block_device(dev):
                raise ValueError(f"{dev} doesn't look like a block device")
            SBDUtils.compare_device_uuid(dev, compare_node_list)
        utils.detect_duplicate_device_path(dev_list)

    @staticmethod
    def get_sbd_value_from_config(key):
        '''
        Get value from /etc/sysconfig/sbd
        '''
        return utils.parse_sysconfig(SBDManager.SYSCONFIG_SBD).get(key)

    @staticmethod
    def get_crashdump_watchdog_timeout() -> typing.Optional[int]:
        res = SBDUtils.get_sbd_value_from_config("SBD_OPTS")
        if not res:
            return None
        matched = re.search(r"-C\s+(\d+)", res)
        return int(matched.group(1)) if matched else None

    @staticmethod
    def get_sbd_device_from_config():
        '''
        Get sbd device list from config
        '''
        res = SBDUtils.get_sbd_value_from_config("SBD_DEVICE")
        return res.split(';') if res else []

    @staticmethod
    def is_using_diskless_sbd():
        '''
        Check if using diskless SBD
        '''
        if not ServiceManager().service_is_active(constants.SBD_SERVICE):
            return False
        return not bool(SBDUtils.get_sbd_device_from_config())

    @staticmethod
    def is_using_disk_based_sbd():
        '''
        Check if using disk-based SBD
        '''
        if not ServiceManager().service_is_active(constants.SBD_SERVICE):
            return False
        return bool(SBDUtils.get_sbd_device_from_config())

    @staticmethod
    def has_sbd_device_already_initialized(dev) -> bool:
        '''
        Check if sbd device already initialized
        '''
        cmd = "sbd -d {} dump".format(dev)
        rc, _, _ = ShellUtils().get_stdout_stderr(cmd)
        return rc == 0

    @staticmethod
    def no_overwrite_device_check(dev) -> bool:
        '''
        Check if device already initialized and ask if need to overwrite
        '''
        initialized = SBDUtils.has_sbd_device_already_initialized(dev)
        return initialized and \
                not bootstrap.confirm(f"{dev} has already been initialized by SBD - overwrite?")

    @staticmethod
    def check_devices_metadata_consistent(dev_list) -> bool:
        '''
        Check if all devices have the same metadata
        '''
        consistent = True
        if len(dev_list) < 2:
            return consistent
        first_dev_metadata = SBDUtils.get_sbd_device_metadata(dev_list[0], timeout_only=True)
        if not first_dev_metadata:
            logger.warning(f"Cannot get metadata for {dev_list[0]}")
            return False
        for dev in dev_list[1:]:
            this_dev_metadata = SBDUtils.get_sbd_device_metadata(dev, timeout_only=True)
            if not this_dev_metadata:
                logger.warning(f"Cannot get metadata for {dev}")
                return False
            if this_dev_metadata != first_dev_metadata:
                logger.warning(f"Device {dev} doesn't have the same metadata as {dev_list[0]}")
                consistent = False
        return consistent

    @staticmethod
    def handle_input_sbd_devices(dev_list, dev_list_from_config=None):
        '''
        Given a list of devices, split them into two lists:
        - overwrite_list: devices that need to be overwritten
        - no_overwrite_list: devices that don't need to be overwritten

        Raise TerminateSubCommand if the metadata of no_overwrite_list is not consistent
        '''
        no_overwrite_list = dev_list_from_config or []
        overwrite_list = []

        for dev in dev_list:
            if SBDUtils.no_overwrite_device_check(dev):
                no_overwrite_list.append(dev)
            else:
                overwrite_list.append(dev)

        if no_overwrite_list and not SBDUtils.check_devices_metadata_consistent(no_overwrite_list):
            raise utils.TerminateSubCommand

        return overwrite_list, no_overwrite_list


class SBDTimeout(object):
    '''
    Consolidate sbd related timeout methods and constants
    '''
    SBD_WATCHDOG_TIMEOUT_DEFAULT = 5
    SBD_WATCHDOG_TIMEOUT_DEFAULT_S390 = 15
    SBD_WATCHDOG_TIMEOUT_DEFAULT_WITH_QDEVICE = 35
    QDEVICE_SYNC_TIMEOUT_MARGIN = 5
    SHOW_SBD_START_TIMEOUT_CMD = "systemctl show -p TimeoutStartUSec sbd.service --value"

    def __init__(self, context=None):
        '''
        Init function
        '''
        self.context = context
        self.sbd_msgwait = None
        self.stonith_timeout = None
        self.sbd_watchdog_timeout = self.SBD_WATCHDOG_TIMEOUT_DEFAULT
        self.stonith_watchdog_timeout = None
        self.two_node_without_qdevice = False

    def initialize_timeout(self):
        self._set_sbd_watchdog_timeout()
        if self.context.diskless_sbd:
            self._adjust_sbd_watchdog_timeout_with_diskless_and_qdevice()
        else:
            self._set_sbd_msgwait()

    def _set_sbd_watchdog_timeout(self):
        '''
        Set sbd_watchdog_timeout from profiles.yml if exists
        Then adjust it if in s390 environment
        '''
        if "sbd.watchdog_timeout" in self.context.profiles_dict:
            self.sbd_watchdog_timeout = int(self.context.profiles_dict["sbd.watchdog_timeout"])
        if self.context.is_s390 and self.sbd_watchdog_timeout < self.SBD_WATCHDOG_TIMEOUT_DEFAULT_S390:
            logger.warning("sbd_watchdog_timeout is set to %d for s390, it was %d", self.SBD_WATCHDOG_TIMEOUT_DEFAULT_S390, self.sbd_watchdog_timeout)
            self.sbd_watchdog_timeout = self.SBD_WATCHDOG_TIMEOUT_DEFAULT_S390

    def _set_sbd_msgwait(self):
        '''
        Set sbd msgwait from profiles.yml if exists
        Default is 2 * sbd_watchdog_timeout
        '''
        sbd_msgwait_default = 2 * self.sbd_watchdog_timeout
        sbd_msgwait = sbd_msgwait_default
        if "sbd.msgwait" in self.context.profiles_dict:
            sbd_msgwait = int(self.context.profiles_dict["sbd.msgwait"])
            if sbd_msgwait < sbd_msgwait_default:
                logger.warning("sbd msgwait is set to %d, it was %d", sbd_msgwait_default, sbd_msgwait)
                sbd_msgwait = sbd_msgwait_default
        self.sbd_msgwait = sbd_msgwait

    @classmethod
    def get_advised_sbd_timeout(cls, diskless=False) -> typing.Tuple[int, int]:
        '''
        Get suitable sbd_watchdog_timeout and sbd_msgwait
        '''
        ctx = bootstrap.Context()
        ctx.diskless_sbd = diskless
        ctx.load_profiles()
        time_inst = cls(ctx)
        time_inst.initialize_timeout()

        sbd_watchdog_timeout = time_inst.sbd_watchdog_timeout
        sbd_msgwait = None if diskless else time_inst.sbd_msgwait
        return sbd_watchdog_timeout, sbd_msgwait

    def _adjust_sbd_watchdog_timeout_with_diskless_and_qdevice(self):
        '''
        When using diskless SBD with Qdevice, adjust value of sbd_watchdog_timeout
        '''
        # add sbd after qdevice started
        if corosync.is_qdevice_configured() and ServiceManager().service_is_active("corosync-qdevice.service"):
            qdevice_sync_timeout = utils.get_qdevice_sync_timeout()
            if self.sbd_watchdog_timeout <= qdevice_sync_timeout:
                watchdog_timeout_with_qdevice = qdevice_sync_timeout + self.QDEVICE_SYNC_TIMEOUT_MARGIN
                logger.warning("sbd_watchdog_timeout is set to {} for qdevice, it was {}".format(watchdog_timeout_with_qdevice, self.sbd_watchdog_timeout))
                self.sbd_watchdog_timeout = watchdog_timeout_with_qdevice
        # add sbd and qdevice together from beginning
        elif self.context.qdevice_inst:
            if self.sbd_watchdog_timeout < self.SBD_WATCHDOG_TIMEOUT_DEFAULT_WITH_QDEVICE:
                logger.warning("sbd_watchdog_timeout is set to {} for qdevice, it was {}".format(self.SBD_WATCHDOG_TIMEOUT_DEFAULT_WITH_QDEVICE, self.sbd_watchdog_timeout))
                self.sbd_watchdog_timeout = self.SBD_WATCHDOG_TIMEOUT_DEFAULT_WITH_QDEVICE

    @staticmethod
    def get_sbd_msgwait(dev):
        '''
        Get msgwait for sbd device
        '''
        res = SBDUtils.get_sbd_device_metadata(dev).get("msgwait")
        if not res:
            raise ValueError(f"Cannot get sbd msgwait for {dev}")
        return res

    @staticmethod
    def get_sbd_watchdog_timeout():
        '''
        Get SBD_WATCHDOG_TIMEOUT from /etc/sysconfig/sbd
        '''
        res = SBDUtils.get_sbd_value_from_config("SBD_WATCHDOG_TIMEOUT")
        if not res:
            raise ValueError("Cannot get the value of SBD_WATCHDOG_TIMEOUT")
        return int(res)

    @staticmethod
    def get_stonith_watchdog_timeout_expected():
        '''
        Returns the value of the stonith-watchdog-timeout cluster property.

        If the Pacemaker service is inactive, returns the default value (2 * SBD_WATCHDOG_TIMEOUT).
        If the property is set and its value is equal to or greater than the default, returns the property value.
        Otherwise, returns the default value.
        '''
        default = 2 * SBDTimeout.get_sbd_watchdog_timeout()
        if not ServiceManager().service_is_active(constants.PCMK_SERVICE):
            return default
        value = utils.get_property("stonith-watchdog-timeout", get_default=False)
        return_value = value if utils.crm_msec(value) >= utils.crm_msec(default) else default
        return int(utils.crm_msec(return_value)/1000)  # convert msec to sec

    def _load_configurations(self):
        '''
        Load necessary configurations for both disk-based/disk-less sbd
        '''
        self.two_node_without_qdevice = utils.is_2node_cluster_without_qdevice()

        dev_list = SBDUtils.get_sbd_device_from_config()
        if dev_list:  # disk-based
            self.disk_based = True
            self.msgwait = SBDTimeout.get_sbd_msgwait(dev_list[0])
            self.pcmk_delay_max = utils.get_pcmk_delay_max(self.two_node_without_qdevice)
        else:  # disk-less
            self.disk_based = False
            self.sbd_watchdog_timeout = SBDTimeout.get_sbd_watchdog_timeout()
            self.stonith_watchdog_timeout = SBDTimeout.get_stonith_watchdog_timeout_expected()
        self.sbd_delay_start_value_expected = self.get_sbd_delay_start_expected() if utils.detect_virt() else "no"
        self.sbd_delay_start_value_from_config = SBDUtils.get_sbd_value_from_config("SBD_DELAY_START")

        logger.debug("Inspect SBDTimeout: %s", vars(self))

    def get_stonith_timeout_expected(self):
        '''
        Get stonith-timeout value for sbd cases, formulas are:

        value_from_sbd = 1.2 * msgwait # for disk-based sbd
        value_from_sbd = 1.2 * max (stonith_watchdog_timeout, 2*SBD_WATCHDOG_TIMEOUT) # for disk-less sbd

        stonith_timeout = max(value_from_sbd, constants.STONITH_TIMEOUT_DEFAULT) + token + consensus
        '''
        if self.disk_based:
            value_from_sbd = int(1.2*self.msgwait)
        else:
            value_from_sbd = int(1.2*max(self.stonith_watchdog_timeout, 2*self.sbd_watchdog_timeout))

        value = max(value_from_sbd, constants.STONITH_TIMEOUT_DEFAULT) + corosync.token_and_consensus_timeout()
        logger.debug("Result of SBDTimeout.get_stonith_timeout_expected %d", value)
        return value

    @classmethod
    def get_stonith_timeout(cls):
        cls_inst = cls()
        cls_inst._load_configurations()
        return cls_inst.get_stonith_timeout_expected()

    def get_sbd_delay_start_expected(self):
        '''
        Get the value for SBD_DELAY_START, formulas are:

        SBD_DELAY_START = (token + consensus + pcmk_delay_max + msgwait)  # for disk-based sbd
        SBD_DELAY_START = (token + consensus + 2*SBD_WATCHDOG_TIMEOUT) # for disk-less sbd
        '''
        token_and_consensus_timeout = corosync.token_and_consensus_timeout()
        if self.disk_based:
            value = token_and_consensus_timeout + self.pcmk_delay_max + self.msgwait
        else:
            value = token_and_consensus_timeout + 2*self.sbd_watchdog_timeout
        return value

    @staticmethod
    def get_sbd_delay_start_sec_from_sysconfig():
        '''
        Get suitable systemd start timeout for sbd.service
        '''
        # TODO 5ms, 5us, 5s, 5m, 5h are also valid for sbd sysconfig
        value = SBDUtils.get_sbd_value_from_config("SBD_DELAY_START")
        if utils.is_boolean_true(value):
            return 2*SBDTimeout.get_sbd_watchdog_timeout()
        return int(value)

    @staticmethod
    def is_sbd_delay_start():
        '''
        Check if SBD_DELAY_START is not no or not set
        '''
        res = SBDUtils.get_sbd_value_from_config("SBD_DELAY_START")
        return res and res != "no"

    @staticmethod
    def get_sbd_systemd_start_timeout() -> int:
        out = sh.cluster_shell().get_stdout_or_raise_error(SBDTimeout.SHOW_SBD_START_TIMEOUT_CMD)
        return utils.get_systemd_timeout_start_in_sec(out)

    def adjust_systemd_start_timeout(self):
        '''
        Adjust start timeout for sbd when set SBD_DELAY_START
        '''
        sbd_delay_start_value = SBDUtils.get_sbd_value_from_config("SBD_DELAY_START")
        if sbd_delay_start_value == "no":
            return

        start_timeout = SBDTimeout.get_sbd_systemd_start_timeout()
        if start_timeout > int(sbd_delay_start_value):
            return

        utils.mkdirp(SBDManager.SBD_SYSTEMD_DELAY_START_DIR)
        sbd_delay_start_file = "{}/sbd_delay_start.conf".format(SBDManager.SBD_SYSTEMD_DELAY_START_DIR)
        utils.str2file("[Service]\nTimeoutSec={}".format(int(1.2*int(sbd_delay_start_value))), sbd_delay_start_file)
        bootstrap.sync_file(SBDManager.SBD_SYSTEMD_DELAY_START_DIR)
        utils.cluster_run_cmd("systemctl daemon-reload")

    def adjust_stonith_timeout(self):
        '''
        Adjust stonith-timeout property
        '''
        utils.set_property("stonith-timeout", self.get_stonith_timeout_expected(), conditional=True)

    def adjust_sbd_delay_start(self):
        '''
        Adjust SBD_DELAY_START in /etc/sysconfig/sbd
        '''
        expected_value = str(self.sbd_delay_start_value_expected)
        config_value = self.sbd_delay_start_value_from_config
        if expected_value == config_value:
            return
        if expected_value == "no" \
                or (not re.search(r'\d+', config_value)) \
                or (int(expected_value) > int(config_value)):
            SBDManager.update_sbd_configuration({"SBD_DELAY_START": expected_value})

    @classmethod
    def adjust_sbd_timeout_related_cluster_configuration(cls):
        '''
        Adjust sbd timeout related configurations
        '''
        cls_inst = cls()
        cls_inst._load_configurations()
        cls_inst.adjust_sbd_delay_start()
        cls_inst.adjust_stonith_timeout()
        cls_inst.adjust_systemd_start_timeout()


class SBDManager:
    SYSCONFIG_SBD = "/etc/sysconfig/sbd"
    SYSCONFIG_SBD_TEMPLATE = "/usr/share/fillup-templates/sysconfig.sbd"
    SBD_SYSTEMD_DELAY_START_DIR = "/etc/systemd/system/sbd.service.d"
    SBD_SYSTEMD_DELAY_START_DISABLE_DIR = "/run/systemd/system/sbd.service.d"
    SBD_SYSTEMD_DELAY_START_DISABLE_FILE = f"{SBD_SYSTEMD_DELAY_START_DISABLE_DIR}/sbd_delay_start_disabled.conf"
    SBD_STATUS_DESCRIPTION = '''Configure SBD:
  If you have shared storage, for example a SAN or iSCSI target,
  you can use it avoid split-brain scenarios by configuring SBD.
  This requires a 1 MB partition, accessible to all nodes in the
  cluster.  The device path must be persistent and consistent
  across all nodes in the cluster, so /dev/disk/by-id/* devices
  are a good choice.  Note that all data on the partition you
  specify here will be destroyed.
'''
    NO_SBD_WARNING = "Not configuring SBD - STONITH will be disabled."
    DISKLESS_SBD_MIN_EXPECTED_VOTE = 3
    DISKLESS_SBD_WARNING = "Diskless SBD requires cluster with three or more nodes. If you want to use diskless SBD for 2-node cluster, should be combined with QDevice."
    SBD_NOT_INSTALLED_MSG = "Package sbd is not installed"
    FENCE_SBD_NOT_EXISTED_MSG = "fence_sbd command does not exist."
    SBD_RA = "stonith:fence_sbd"
    SBD_RA_ID = "stonith-sbd"
    SBD_DEVICE_MAX = 3

    class NotConfigSBD(Exception):
        pass

    def __init__(
        self,
        device_list_to_init: typing.List[str] | None = None,
        timeout_dict: typing.Dict[str, int] | None = None,
        update_dict: typing.Dict[str, str] | None = None,
        diskless_sbd: bool = False,
        bootstrap_context: 'bootstrap.Context | None' = None
    ):
        '''
        Init function which can be called from crm sbd subcommand or bootstrap
        '''
        self.device_list_to_init = device_list_to_init or []
        self.timeout_dict = timeout_dict or {}
        self.update_dict = update_dict or {}
        self.diskless_sbd = diskless_sbd
        self.cluster_is_running = ServiceManager().service_is_active(constants.PCMK_SERVICE)
        self.bootstrap_context = bootstrap_context
        self.overwrite_sysconfig = False

        # From bootstrap init or join process, override the values
        if self.bootstrap_context:
            self.overwrite_sysconfig = self.bootstrap_context.type == "init"
            self.diskless_sbd = self.bootstrap_context.diskless_sbd
            self.cluster_is_running = self.bootstrap_context.cluster_is_running

    def _load_attributes_from_bootstrap(self):
        if not self.bootstrap_context or not self.overwrite_sysconfig:
            return
        if not self.timeout_dict:
            timeout_inst = SBDTimeout(self.bootstrap_context)
            timeout_inst.initialize_timeout()
            self.timeout_dict["watchdog"] = timeout_inst.sbd_watchdog_timeout
            if self.diskless_sbd:
                self.update_dict["SBD_WATCHDOG_TIMEOUT"] = str(timeout_inst.sbd_watchdog_timeout)
            else:
                self.timeout_dict["msgwait"] = timeout_inst.sbd_msgwait
        self.update_dict["SBD_WATCHDOG_DEV"] = watchdog.Watchdog.get_watchdog_device(self.bootstrap_context.watchdog)

    @staticmethod
    def convert_timeout_dict_to_opt_str(timeout_dict: typing.Dict[str, int]) -> str:
        timeout_option_map = {
            "watchdog": "-1",
            "allocate": "-2",
            "loop": "-3",
            "msgwait": "-4"
        }
        return ' '.join([f"{timeout_option_map[k]} {v}" for k, v in timeout_dict.items()
                         if k in timeout_option_map])

    def update_configuration(self) -> None:
        '''
        Update and sync sbd configuration
        '''
        if not self.update_dict:
            return
        if self.overwrite_sysconfig:
            utils.copy_local_file(self.SYSCONFIG_SBD_TEMPLATE, self.SYSCONFIG_SBD)

        for key, value in self.update_dict.items():
            logger.info("Update %s in %s: %s", key, self.SYSCONFIG_SBD, value)
        utils.sysconfig_set(self.SYSCONFIG_SBD, **self.update_dict)
        bootstrap.sync_file(self.SYSCONFIG_SBD)
        logger.info("Already synced %s to all nodes", self.SYSCONFIG_SBD)

    @classmethod
    def update_sbd_configuration(cls, update_dict: typing.Dict[str, str]) -> None:
        inst = cls(update_dict=update_dict)
        inst.update_configuration()

    def initialize_sbd(self):
        if self.diskless_sbd:
            logger.info("Configuring diskless SBD")
            self._warn_diskless_sbd()
            return
        elif self.device_list_to_init:
            logger.info("Configuring disk-based SBD")
        else:
            return

        if not shutil.which("fence_sbd"):
            utils.fatal(self.FENCE_SBD_NOT_EXISTED_MSG)

        opt_str = SBDManager.convert_timeout_dict_to_opt_str(self.timeout_dict)
        shell = sh.cluster_shell()
        for dev in self.device_list_to_init:
            logger.info("Initializing SBD device %s", dev)
            cmd = f"sbd {opt_str} -d {dev} create"
            logger.debug("Running command: %s", cmd)
            shell.get_stdout_or_raise_error(cmd)

    @staticmethod
    def enable_sbd_service():
        cluster_nodes = utils.list_cluster_nodes() or [utils.this_node()]
        service_manager = ServiceManager()

        for node in cluster_nodes:
            if not service_manager.service_is_enabled(constants.SBD_SERVICE, node):
                logger.info("Enable %s on node %s", constants.SBD_SERVICE, node)
                service_manager.enable_service(constants.SBD_SERVICE, node)

    @staticmethod
    def restart_cluster_if_possible(with_maintenance_mode=False):
        if not ServiceManager().service_is_active(constants.PCMK_SERVICE):
            return
        if not xmlutil.CrmMonXmlParser().is_non_stonith_resource_running():
            bootstrap.restart_cluster()
        elif with_maintenance_mode:
            if not utils.is_dlm_running():
                bootstrap.restart_cluster()
            else:
                logger.warning("Resource is running, need to restart cluster service manually on each node")
        else:
            logger.warning("Resource is running, need to restart cluster service manually on each node")
            logger.warning("Or, run with `crm -F` or `--force` option, the `sbd` subcommand will leverage maintenance mode for any changes that require restarting sbd.service")
            logger.warning("Understand risks that running RA has no cluster protection while the cluster is in maintenance mode and restarting")

    def configure_sbd(self):
        '''
        Configure fence_sbd resource and related properties
        '''
        if SBDUtils.get_sbd_device_from_config():
            if utils.get_property("stonith-watchdog-timeout", get_default=False):
                utils.delete_property("stonith-watchdog-timeout")
            if not xmlutil.CrmMonXmlParser().is_resource_configured(self.SBD_RA):
                cmd = f"crm configure primitive {self.SBD_RA_ID} {self.SBD_RA}"
                sh.cluster_shell().get_stdout_or_raise_error(cmd)
        else:
            swt_value = self.timeout_dict.get("stonith-watchdog", SBDTimeout.get_stonith_watchdog_timeout_expected())
            utils.set_property("stonith-watchdog-timeout", swt_value)
        utils.set_property("stonith-enabled", "true")

    def _warn_diskless_sbd(self, peer=None):
        '''
        Give warning when configuring diskless sbd
        '''
        # When in sbd stage or join process
        if (self.diskless_sbd and self.cluster_is_running) or peer:
            vote_dict = utils.get_quorum_votes_dict(peer)
            expected_vote = int(vote_dict.get('Expected', 0))
            if expected_vote < self.DISKLESS_SBD_MIN_EXPECTED_VOTE:
                logger.warning('%s', self.DISKLESS_SBD_WARNING)
        # When in init process
        elif self.diskless_sbd:
            logger.warning('%s', self.DISKLESS_SBD_WARNING)

    def _warn_and_raise_no_sbd(self):
        logger.warning('%s', self.NO_SBD_WARNING)
        raise self.NotConfigSBD

    def _wants_to_overwrite(self, configured_devices):
        wants_to_overwrite_msg = f"SBD_DEVICE in {self.SYSCONFIG_SBD} is already configured to use '{';'.join(configured_devices)}' - overwrite?"
        if not bootstrap.confirm(wants_to_overwrite_msg):
            if not SBDUtils.check_devices_metadata_consistent(configured_devices):
                raise utils.TerminateSubCommand
            self.overwrite_sysconfig = False
            return False
        return True

    def _prompt_for_sbd_device(self) -> list[str]:
        '''
        Prompt for sbd device and verify
        '''
        dev_list = []
        dev_looks_sane = False
        while not dev_looks_sane:
            dev = bootstrap.prompt_for_string('Path to storage device (e.g. /dev/disk/by-id/...), or "none" for diskless sbd, use ";" as separator for multi path', r'none|\/.*')
            if dev == "none":
                self.diskless_sbd = True
                return []

            dev_list = utils.re_split_string("[; ]", dev)
            try:
                SBDUtils.verify_sbd_device(dev_list)
            except ValueError as e:
                logger.error('%s', e)
                continue
            for dev in dev_list:
                if SBDUtils.has_sbd_device_already_initialized(dev):
                    dev_looks_sane = True
                    continue
                else:
                    logger.warning("All data on %s will be destroyed", dev)
                    if bootstrap.confirm('Are you sure you wish to use this device?'):
                        dev_looks_sane = True
                    else:
                        dev_looks_sane = False
                        break
        return dev_list

    def get_sbd_device_interactive(self) -> list[str]:
        '''
        Get sbd device on interactive mode
        '''
        if self.bootstrap_context.yes_to_all:
            self._warn_and_raise_no_sbd()
        logger.info(self.SBD_STATUS_DESCRIPTION)
        if not bootstrap.confirm("Do you wish to use SBD?"):
            self._warn_and_raise_no_sbd()
        if not utils.package_is_installed("sbd"):
            utils.fatal(self.SBD_NOT_INSTALLED_MSG)

        configured_devices = SBDUtils.get_sbd_device_from_config()
        # return empty list if already configured and user doesn't want to overwrite
        if configured_devices and not self._wants_to_overwrite(configured_devices):
            return []

        return self._prompt_for_sbd_device()

    def get_sbd_device_from_bootstrap(self):
        '''
        Handle sbd device input from 'crm cluster init' with -s or -S option
        -s is for disk-based sbd
        -S is for diskless sbd
        '''
        # if specified sbd device with -s option
        device_list = self.bootstrap_context.sbd_devices
        # else if not use -S option, get sbd device interactively
        if not device_list and not self.bootstrap_context.diskless_sbd:
            device_list = self.get_sbd_device_interactive()
        if not device_list:
            return

        # get two lists of devices, one for overwrite, one for no overwrite with consistent metadata
        overwrite_list, no_overwrite_list = SBDUtils.handle_input_sbd_devices(device_list)
        self.device_list_to_init = overwrite_list
        # if no_overwrite_list is not empty, get timeout metadata from the first device
        if no_overwrite_list:
            self.timeout_dict = SBDUtils.get_sbd_device_metadata(no_overwrite_list[0], timeout_only=True)
        self.update_dict["SBD_DEVICE"] = ';'.join(device_list)

    def init_and_deploy_sbd(self):
        '''
        The process of deploying sbd includes:
        1. Initialize sbd device
        2. Write config file /etc/sysconfig/sbd
        3. Enable sbd.service
        4. Restart cluster service if possible
        5. Configure stonith-sbd resource and related properties
        '''
        if self.bootstrap_context:
            try:
                self.get_sbd_device_from_bootstrap()
            except self.NotConfigSBD:
                ServiceManager().disable_service(constants.SBD_SERVICE)
                return
            self._load_attributes_from_bootstrap()

        with utils.leverage_maintenance_mode() as enabled:
            self.initialize_sbd()
            self.update_configuration()
            SBDManager.enable_sbd_service()

            if self.cluster_is_running:
                self.configure_sbd()
                bootstrap.adjust_properties(with_sbd=True)
                SBDManager.restart_cluster_if_possible(with_maintenance_mode=enabled)

    def join_sbd(self, remote_user, peer_host):
        '''
        Function join_sbd running on join process only
        On joining process, check whether peer node has enabled sbd.service
        If so, check prerequisites of SBD and verify sbd device on join node
        '''
        service_manager = ServiceManager()
        if not os.path.exists(self.SYSCONFIG_SBD) or not service_manager.service_is_enabled(constants.SBD_SERVICE, peer_host):
            service_manager.disable_service(constants.SBD_SERVICE)
            return

        from .watchdog import Watchdog
        self._watchdog_inst = Watchdog(remote_user=remote_user, peer_host=peer_host)
        self._watchdog_inst.join_watchdog()

        dev_list = SBDUtils.get_sbd_device_from_config()
        if dev_list:
            SBDUtils.verify_sbd_device(dev_list, [peer_host])
        else:
            self._warn_diskless_sbd(peer_host)

        logger.info("Got {}SBD configuration".format("" if dev_list else "diskless "))
        service_manager.enable_service(constants.SBD_SERVICE)


def cleanup_existing_sbd_resource():
    if xmlutil.CrmMonXmlParser().is_resource_configured(SBDManager.SBD_RA):
        sbd_id_list = xmlutil.CrmMonXmlParser().get_resource_id_list_via_type(SBDManager.SBD_RA)
        if xmlutil.CrmMonXmlParser().is_resource_started(SBDManager.SBD_RA):
            for sbd_id in sbd_id_list:
                logger.info("Stop sbd resource '%s'(%s)", sbd_id, SBDManager.SBD_RA)
                utils.ext_cmd("crm resource stop {}".format(sbd_id))
        logger.info("Remove sbd resource '%s'", ';' .join(sbd_id_list))
        utils.ext_cmd("crm configure delete {}".format(' '.join(sbd_id_list)))


def purge_sbd_from_cluster():
    '''
    Purge SBD from cluster, the process includes:
    - stop and remove sbd agent
    - disable sbd.service
    - move /etc/sysconfig/sbd to /etc/sysconfig/sbd.bak
    - adjust cluster attributes
    - adjust related timeout values
    '''
    cleanup_existing_sbd_resource()

    cluster_nodes = utils.list_cluster_nodes()
    service_manager = ServiceManager()
    for node in cluster_nodes:
        if service_manager.service_is_enabled(constants.SBD_SERVICE, node):
            logger.info("Disable %s on node %s", constants.SBD_SERVICE, node)
            service_manager.disable_service(constants.SBD_SERVICE, node)

    config_bak = f"{SBDManager.SYSCONFIG_SBD}.bak"
    logger.info("Move %s to %s on all nodes", SBDManager.SYSCONFIG_SBD, config_bak)
    utils.cluster_run_cmd(f"mv {SBDManager.SYSCONFIG_SBD} {config_bak}")

    out = sh.cluster_shell().get_stdout_or_raise_error("stonith_admin -L")
    res = re.search("([0-9]+) fence device[s]* found", out)
    # after disable sbd.service, check if sbd is the last stonith device
    if res and int(res.group(1)) <= 1:
        utils.cleanup_stonith_related_properties()

    for _dir in [SBDManager.SBD_SYSTEMD_DELAY_START_DIR, SBDManager.SBD_SYSTEMD_DELAY_START_DISABLE_DIR]:
        cmd = f"test -d {_dir} && rm -rf {_dir} || exit 0"
        parallax.parallax_call(cluster_nodes, cmd)



# Copyright (C) 2008-2011 Dejan Muhamedagic <dmuhamedagic@suse.de>
# Copyright (C) 2013 Kristoffer Gronlund <kgronlund@suse.com>
# See COPYING for license information.

'''
The commands exposed by this module all
get their data from the doc/crm.8.adoc text
file. In that file, there are help for
 - topics
 - levels
 - commands in levels

The help file is lazily loaded when the first
request for help is made.

All help is in the following form in the manual:
[[cmdhelp_<level>_<cmd>,<short help text>]]
=== ...
Long help text.
...
[[cmdhelp_<level>_<cmd>,<short help text>]]

Help for the level itself is like this:

[[cmdhelp_<level>,<short help text>]]
'''
import dataclasses
import functools
import io
import os
import re
import typing

from .sh import ShellUtils
from .utils import fuzzy_get
from .utils import page_string
from . import config
from . import clidisplay
from . import log
from . import ra


logger = log.setup_logger(__name__)


class HelpFilter(object):
    _B0 = re.compile(r'^\.{4,}')
    _B1 = re.compile(r'^\*{4,}')
    _QUOTED = re.compile(r'`([^`]+)`')
    _MONO = re.compile(r'\+([^+]+)\+')
    _TOPIC = re.compile(r'(.*)::$')
    _TOPIC2 = re.compile(r'^\.\w+')

    def __init__(self):
        self.in_block = False

    def _filter(self, line):
        block_edge = self._B0.match(line) or self._B1.match(line)
        if block_edge and not self.in_block:
            self.in_block = True
            return ''
        elif block_edge and self.in_block:
            self.in_block = False
            return ''
        elif not self.in_block:
            if self._TOPIC2.match(line):
                return clidisplay.help_topic(line[1:])
            line = self._QUOTED.sub(clidisplay.help_keyword(r'\1'), line)
            line = self._MONO.sub(clidisplay.help_block(r'\1'), line)
            line = self._TOPIC.sub(clidisplay.help_topic(r'\1'), line)
            return line
        else:
            return clidisplay.help_block(line)

    def __call__(self, text):
        return '\n'.join([self._filter(line) for line in text.splitlines()]) + '\n'


class HelpEntry(object):
    def __init__(self, short_help, long_help='', alias_for=None, generated=False):
        if short_help:
            self.short = short_help[0].upper() + short_help[1:]
        else:
            self.short = 'Help'
        self._long_help = long_help
        self.alias_for = alias_for
        self.generated = generated

    @property
    def long_help(self):
        return self._long_help

    @long_help.setter
    def long_help(self, value):
        self._long_help = value

    def is_alias(self):
        return self.alias_for is not None

    def paginate(self):
        '''
        Display help, paginated.
        Replace asciidoc syntax with colorized output where possible.
        '''
        helpfilter = HelpFilter()

        short_help = clidisplay.help_header(self.short)
        long_help = self.long_help
        if long_help:
            long_help = helpfilter(long_help)
            if not long_help.startswith('\n'):
                long_help = '\n' + long_help

        prefix = ''
        if self.is_alias():
            prefix = helpfilter("(Redirected from `%s` to `%s`)\n" % self.alias_for)

        page_string(short_help + '\n' + prefix + long_help)

    def __str__(self):
        if self.long_help:
            return self.short + '\n' + self.long_help
        return self.short

    def __repr__(self):
        return str(self)


class LazyHelpEntryFromCli(HelpEntry):
    """lazy load from cli --help"""
    def __init__(
            self,
            short_help: str,
            cmd_args: typing.Sequence[str],
            alias_for=None, generated=False,
    ):
        super().__init__(short_help, '', alias_for, generated)
        self._cmd_args = cmd_args

    @functools.cached_property
    def long_help(self):
        args = ['crm']
        args.extend(self._cmd_args)
        args.append('--help-without-redirect')
        rc, stdout, _ = ShellUtils.get_stdout_stderr(args, shell=False, mix_stderr=True)
        return stdout

    def paginate(self):
        page_string('{}\n\n{}'.format(self.short, self.long_help))


@dataclasses.dataclass
class SubcommandTreeNode:
    """Tree node representing subcommmand hierarchy and holding HelpEntries"""
    name: str
    help: HelpEntry
    children: dict[str, typing.Self]

    def make_sublevel_recursively(self, sublevels: typing.Sequence[str]) -> typing.Self:
        current_level = self
        for sublevel in sublevels:
            x = current_level.children.get(sublevel, None)
            if x is None:
                x = SubcommandTreeNode(sublevel, HelpEntry(''), dict())
                current_level.children[sublevel] = x
            current_level = x
        return current_level


HELP_FILE = os.path.join(config.path.sharedir, 'crm.8.adoc')

_DEFAULT = HelpEntry('No help available', long_help='', alias_for=None, generated=True)
_REFERENCE_RE = re.compile(r'<<[^,]+,(.+)>>')

# loaded on demand
# _LOADED is set to True when an attempt
# has been made (so it won't be tried again)
_LOADED = False
_TOPICS = {}
_COMMAND_TREE = SubcommandTreeNode(
    'crm',
    HelpEntry(''),
    dict()
)

_TOPICS["Overview"] = HelpEntry("Available help topics and commands", generated=True)
_TOPICS["Topics"] = HelpEntry("Available help topics", generated=True)


def _titleline(title, desc, suffix='', width=16):
    return '%-0*s %s\n' % (width, ('`%s`' % (title)) + suffix, desc)


_hidden_commands = ('up', 'cd', 'help', 'quit', 'ls')


def get_max_width(dict_):
    max_width = 16
    for key in list(dict_.keys()):
        if max_width < len(key):
            max_width = len(key)
    if max_width >= 16:
        max_width += 2
    return max_width


def _render_command_tree(out: io.StringIO, node: SubcommandTreeNode, indent: int):
    max_width = get_max_width(node.children)
    for name, node in sorted(node.children.items(), key=lambda x: (bool(x[1].children), x[0])):
        if name == node.name:   # is not an alias
            for _ in range(indent):
                out.write('\t')
            if node.children:
                out.write(_titleline(name, node.help.short, suffix='/', width=max_width))
                _render_command_tree(out, node, indent + 1)
                out.write('\n')
            else:
                out.write(_titleline(name, node.help.short, width=max_width))


def help_overview():
    '''
    Returns an overview of all available
    topics and commands.
    '''
    _load_help()
    s = io.StringIO()
    s.write("Available topics:\n\n")
    max_width = get_max_width(_TOPICS)
    for title, topic in _TOPICS.items():
        s.write('\t' + _titleline(title, topic.short, width=max_width))
    s.write("\nAvailable commands:\n\n")
    _render_command_tree(s, _COMMAND_TREE, 1)
    return HelpEntry('Help overview for crmsh\n', s.getvalue(), generated=True)


def help_topics():
    '''
    Returns an overview of all available
    topics.
    '''
    _load_help()
    s = ''
    max_width = get_max_width(_TOPICS)
    for title, topic in _TOPICS.items():
        s += '\t' + _titleline(title, topic.short, width=max_width)
    return HelpEntry('Available topics\n', s, generated=True)


def list_help_topics():
    _load_help()
    return list(_TOPICS.keys())


def help_topic(topic):
    '''
    Returns a help entry for a given topic.
    '''
    _load_help()
    return _TOPICS.get(topic, _DEFAULT)


def _get_level_help(root: SubcommandTreeNode, levels: typing.Sequence[str]):
    node = root
    for level in levels:
        if node is not None:
            node = fuzzy_get(node.children, level)
        else:
            return None
    return node.help


def help_level(levels: typing.Sequence[str]):
    '''
    Returns a help entry for a given level.
    '''
    _load_help()
    help = _get_level_help(_COMMAND_TREE, levels)
    if help is None:
        return _DEFAULT
    else:
        return help


def help_command(levels: typing.Sequence[str]):
    '''
    Returns a help entry for a given command
    '''
    _load_help()
    help = _get_level_help(_COMMAND_TREE, levels)
    if help is None:
            raise ValueError("Undocumented topic '{}'".format(' '.join(levels)))
    else:
        return help


def _is_help_topic(arg):
    return arg and arg[0].isupper()


def _is_command(levels: typing.Sequence[str]):
    node = _COMMAND_TREE
    for level in levels:
        if node is not None:
            node = fuzzy_get(node.children, level)
    return node is not None and not node.children


def _is_level(levels: typing.Sequence[str]):
    node = _COMMAND_TREE
    for level in levels:
        if node is not None:
            node = fuzzy_get(node.children, level)
    return node is not None and node.children


def _is_property(levels: typing.Sequence[str]):
    return len(levels) == 3 and levels[0] == 'configure' and levels[1] == 'property'


def help_contextual(levels: typing.Sequence[str]):
    """
    Returns contextual help
    """
    _load_help()
    if len(levels) == 0:
        return help_overview()
    elif len(levels) == 1 and _is_help_topic(levels[0]):
        return help_topic(levels[0])
    elif _is_command(levels) or _is_level(levels):
        return help_command(levels[0:])
    elif _is_property(levels):
        agent = ra.get_properties_meta()
        if agent:
            all_properties = agent.params()
            property_name = levels[-1]
            if property_name in all_properties:
                print(agent.meta_parameter(property_name))
            else:
                raise ValueError(f"Unknown property '{property_name}'")
        return None
    else:
        topic = levels[0].lower()
        t = fuzzy_get(_TOPICS, topic)
        if t:
            return t
    raise ValueError('No help found for "crm {}". Run "crm help overview" to list all help entries.'.format(
        ' '.join(levels),
    ))


def add_help(levels: typing.Sequence[str], entry):
    '''
    Takes a help entry as argument and inserts it into the
    help system.

    Used to define some help texts statically, for example
    for 'up' and 'help' itself.
    '''
    node = _COMMAND_TREE.make_sublevel_recursively(levels[:-1])
    node.children[levels[-1]] = SubcommandTreeNode(levels[-1], entry, dict())


def _load_help():
    '''
    Lazily load and parse crm.8.adoc.
    '''
    global _LOADED
    if _LOADED:
        return
    _LOADED = True

    def parse_header(line):
        'returns a new entry'
        entry = {'type': '', 'name': '', 'short': '', 'long': '', "from_cli": False}
        line = line[2:-3]  # strip [[ and ]]\n
        info, short_help = line.split(',', 1)
        # TODO see https://github.com/ClusterLabs/crmsh/pull/644
        # This solution has shortcome to delete the content of adoc,
        # which lose the static man page archive
        if "From Code" in short_help:
            short_help, _ = short_help.split(',')
            entry['from_cli'] = True
        entry['short'] = short_help.strip()
        info = info.split('.')
        if info[0] == 'topics':
            entry['type'] = 'topic'
            entry['name'] = info[-1]
        elif info[0] == 'cmdhelp':
            if len(info) == 2:
                entry['type'] = 'level'
                entry['name'] = info[1]
            elif len(info) >= 3:
                entry['type'] = 'command'
                entry['subcommand'] = info[1:]

        return entry

    def process(entry):
        'writes the entry into topics/levels/commands'
        short_help = entry['short']
        long_help = entry['long']
        if long_help.startswith('=='):
            long_help = long_help.split('\n', 1)[1]
        name = entry['name']
        if entry['type'] == 'topic':
            _TOPICS[name] = HelpEntry(short_help, long_help.rstrip())
        elif entry['type'] == 'level':
            _COMMAND_TREE.children[name] = SubcommandTreeNode(name, HelpEntry(short_help, long_help.rstrip()), dict())
        elif entry['type'] == 'command':
            subcommand = entry['subcommand']
            if entry['from_cli']:
                help_entry = LazyHelpEntryFromCli(short_help, subcommand)
            else:
                help_entry = HelpEntry(short_help, long_help.rstrip())
            level = _COMMAND_TREE.make_sublevel_recursively(subcommand[:-1])
            level.children[subcommand[-1]] = SubcommandTreeNode(subcommand[-1], help_entry, dict())

    def filter_line(line):
        '''clean up an input line
         - <<...>> references -> short description
        '''
        return _REFERENCE_RE.sub(r'\1', line)

    def append_subcommand_list_to_long_description(node: SubcommandTreeNode):
        if isinstance(node.help, LazyHelpEntryFromCli) or not node.children:
            return
        buf = io.StringIO()
        buf.write(node.help.long_help)
        buf.write('\n\nCommands:\n')
        max_width = get_max_width(node.children)
        for name, child in sorted(node.children.items(), key=lambda x: (bool(x[1].children), x[0])):
            if name in _hidden_commands or name.startswith('_'):
                continue
            buf.write('\t')
            buf.write(_titleline(name, child.help.short, width=max_width))
            append_subcommand_list_to_long_description(child)
        node.help.long_help = buf.getvalue()

    def fixup_help_aliases(help_node: SubcommandTreeNode, childinfo):
        "add help for aliases"
        for name, childinfo in childinfo.children.items():
            if name in help_node.children:
                fixup_help_aliases(help_node.children[name], childinfo)
                for alias in childinfo.aliases:
                    help_node.children[alias] = help_node.children[name]
        return

    def fixup_topics():
        "fix entries for topics and overview"
        _TOPICS["Overview"] = help_overview()
        _TOPICS["Topics"] = help_topics()

    try:
        name = os.getenv("CRM_HELP_FILE") or HELP_FILE
        with open(name, 'r') as helpfile:
            entry = None
            for line in helpfile:
                if line.startswith('[['):
                    if entry is not None:
                        process(entry)
                    entry = parse_header(line)
                elif entry is not None and line.startswith('===') and entry['long']:
                    process(entry)
                    entry = None
                elif entry is not None:
                    entry['long'] += filter_line(line)
            if entry is not None:
                process(entry)
        append_subcommand_list_to_long_description(_COMMAND_TREE)
        #fixup_root_commands()
        from . import ui_root
        for name, childinfo in ui_root.Root.children().items():
            # the 1st level uses children(), but deeper levels uses children
            # this requires us to write the 1st level of the recursion separately
            if name in _COMMAND_TREE.children:
                fixup_help_aliases(_COMMAND_TREE.children[name], childinfo)
                for alias in childinfo.aliases:
                    ui_root.Root.children()[alias] = ui_root.Root.children()[name]
        fixup_topics()
    except IOError as msg:
        logger.error("Help text not found! %s", msg)

# vim:ts=4:sw=4:et:



import subprocess
import unittest
import pickle
from unittest import mock

import crmsh.sh
from crmsh.user_of_host import UserOfHost


class TestLocalShell(unittest.TestCase):
    def setUp(self) -> None:
        self.local_shell = crmsh.sh.LocalShell()
        self.local_shell.get_effective_user_name = mock.Mock(self.local_shell.get_effective_user_name)
        self.local_shell.geteuid = mock.Mock(self.local_shell.geteuid)
        self.local_shell.hostname = mock.Mock(self.local_shell.hostname)

    @mock.patch('os.environ')
    @mock.patch('subprocess.run')
    def test_su_subprocess_run(self, mock_run: mock.MagicMock, mock_environ: mock.MagicMock):
        self.local_shell.get_effective_user_name.return_value = 'root'
        self.local_shell.geteuid.return_value = 0
        self.local_shell.su_subprocess_run(
            'alice', 'foo',
            input=b'bar',
        )
        mock_run.assert_called_once_with(
            ['su', 'alice', '--login', '-s', '/bin/sh', '-c', 'foo'],
            input=b'bar',
            env=mock_environ,
        )

    @mock.patch('os.environ')
    @mock.patch('subprocess.run')
    def test_su_subprocess_run_as_root(self, mock_run: mock.MagicMock, mock_environ: mock.MagicMock):
        self.local_shell.get_effective_user_name.return_value = 'root'
        self.local_shell.geteuid.return_value = 0
        self.local_shell.su_subprocess_run(
            'root', 'foo',
            input=b'bar',
        )
        mock_run.assert_called_once_with(
            ['/bin/sh', '-c', 'foo'],
            input=b'bar',
            env=mock_environ,
        )

    @mock.patch('subprocess.run')
    def test_su_subprocess_run_unauthorized(self, mock_run: mock.MagicMock):
        self.local_shell.get_effective_user_name.return_value = 'bob'
        self.local_shell.geteuid.return_value = 1001
        with self.assertRaises(crmsh.sh.AuthorizationError) as ctx:
            self.local_shell.su_subprocess_run(
                'root', 'foo',
                input=b'bar',
            )
        self.assertIsInstance(ctx.exception, ValueError)

    def test_get_stdout_stderr_decoded_and_stripped(self):
        self.local_shell.get_rc_stdout_stderr_raw = mock.Mock(self.local_shell.get_rc_stdout_stderr_raw)
        self.local_shell.get_rc_stdout_stderr_raw.return_value = 1, b' out \n', b'\terr\t\n'
        rc, out, err = self.local_shell.get_rc_stdout_stderr('alice', 'foo', 'input')
        self.assertEqual(1, rc)
        self.assertEqual('out', out)
        self.assertEqual('err', err)
        self.local_shell.get_rc_stdout_stderr_raw.assert_called_once_with(
            'alice', 'foo', b'input',
        )

    def test_get_stdout_or_raise_error(self):
        self.local_shell.su_subprocess_run = mock.Mock(self.local_shell.su_subprocess_run)
        self.local_shell.su_subprocess_run.return_value = subprocess.CompletedProcess(
            args=mock.Mock(),
            returncode=1,
            stdout=b'foo',
            stderr=b' bar ',
        )
        with self.assertRaises(crmsh.sh.CommandFailure) as ctx:
            self.local_shell.get_stdout_or_raise_error('root', 'foo')
        self.assertIsInstance(ctx.exception, ValueError)


class TestSSHShell(unittest.TestCase):
    def setUp(self) -> None:
        self.ssh_shell = crmsh.sh.SSHShell(mock.Mock(crmsh.sh.LocalShell), 'alice')
        self.ssh_shell.local_shell.hostname.return_value = 'node1'
        self.ssh_shell.local_shell.get_effective_user_name.return_value = 'root'
        self.ssh_shell.local_shell.can_run_as.return_value = True

    def test_can_run_as(self):
        self.ssh_shell.local_shell.get_rc_and_error.return_value = 255, 'bar'
        self.assertFalse(self.ssh_shell.can_run_as('node2', 'root'))
        self.ssh_shell.local_shell.can_run_as.assert_not_called()

    def test_can_run_as_local(self):
        self.assertTrue(self.ssh_shell.can_run_as(None, 'root'))
        self.ssh_shell.local_shell.can_run_as.assert_called_once_with('root')

    def test_subprocess_run_without_input(self):
        self.ssh_shell.subprocess_run_without_input(
            'node2', 'bob',
            'foo',
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        args, kwargs = self.ssh_shell.local_shell.su_subprocess_run.call_args
        self.assertEqual('alice', args[0])
        self.assertIn('bob@node2', args[1])
        self.assertEqual(b'foo', kwargs['input'])
        self.assertEqual(subprocess.PIPE, kwargs['stdout'])
        self.assertEqual(subprocess.PIPE, kwargs['stderr'])

    def test_subprocess_run_without_input_with_input_kwargs(self):
        with self.assertRaises(AssertionError):
            self.ssh_shell.subprocess_run_without_input(
                'node2', 'bob',
                'foo',
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                input=b'bar'
            )
        self.ssh_shell.local_shell.su_subprocess_run.assert_not_called()
        with self.assertRaises(AssertionError):
            self.ssh_shell.subprocess_run_without_input(
                'node2', 'bob',
                'foo',
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                stdin=subprocess.PIPE,
            )
        self.ssh_shell.local_shell.su_subprocess_run.assert_not_called()

    @mock.patch('os.environ')
    @mock.patch('subprocess.run')
    def test_subprocess_run_without_input_local(self, mock_run, mock_environ: mock.MagicMock):
        self.ssh_shell.subprocess_run_without_input(
            'node1', 'bob',
            'foo',
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        self.ssh_shell.local_shell.su_subprocess_run.assert_not_called()
        mock_run.assert_called_once_with(
            ['sudo', '-H', '-u', 'bob', '/bin/sh'],
            input=b'foo',
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            env=mock_environ,
        )


class TestClusterShell(unittest.TestCase):
    def setUp(self) -> None:
        self.cluster_shell = crmsh.sh.ClusterShell(mock.Mock(crmsh.sh.LocalShell), mock.Mock(UserOfHost))
        self.cluster_shell.local_shell.hostname.return_value = 'node1'
        self.cluster_shell.local_shell.get_effective_user_name.return_value = 'root'
        self.cluster_shell.local_shell.can_run_as.return_value = True
        self.cluster_shell.user_of_host.user_pair_for_ssh.return_value = ('alice', 'bob')

    def test_subprocess_run_without_input(self):
        self.cluster_shell.subprocess_run_without_input(
            'node2',
            None,
            'foo',
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        self.cluster_shell.user_of_host.user_pair_for_ssh.assert_called_once_with('node2')
        args, kwargs = self.cluster_shell.local_shell.su_subprocess_run.call_args
        self.assertEqual('alice', args[0])
        self.assertIn('bob@node2', args[1])
        self.assertIn('-u root', args[1])
        self.assertEqual(b'foo', kwargs['input'])
        self.assertEqual(subprocess.PIPE, kwargs['stdout'])
        self.assertEqual(subprocess.PIPE, kwargs['stderr'])

    def test_subprocess_run_without_input_with_input_kwargs(self):
        with self.assertRaises(AssertionError):
            self.cluster_shell.subprocess_run_without_input(
                'node2',
                None,
                'foo',
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                input=b'bar',
            )
        self.cluster_shell.local_shell.su_subprocess_run.assert_not_called()
        with self.assertRaises(AssertionError):
            self.cluster_shell.subprocess_run_without_input(
                'node2',
                None,
                'foo',
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                stdin=subprocess.PIPE,
            )
        self.cluster_shell.local_shell.su_subprocess_run.assert_not_called()


class TestCommandFailurePickling(unittest.TestCase):
    def test_pickling_unpickling(self):
        # Create an instance of CommandFailure
        original = crmsh.sh.CommandFailure(cmd="ls", host="localhost", user="root", msg="Permission denied")
        # Pickle the object
        pickled = pickle.dumps(original)
        # Unpickle the object
        unpickled = pickle.loads(pickled)

        # Assert that the unpickled object retains the same attributes as the original
        self.assertEqual(original.cmd, unpickled.cmd)
        self.assertEqual(original.host, unpickled.host)
        self.assertEqual(original.user, unpickled.user)
        self.assertEqual(original.msg, unpickled.msg)

    def test_pickling_unpickling_with_none_values(self):
        # Create an instance of CommandFailure with None values for optional parameters
        original = crmsh.sh.CommandFailure(cmd="ls", host=None, user=None, msg="No such file or directory")
        # Pickle the object
        pickled = pickle.dumps(original)
        # Unpickle the object
        unpickled = pickle.loads(pickled)

        # Assert that the unpickled object retains the same attributes, including None values
        self.assertEqual(original.cmd, unpickled.cmd)
        self.assertEqual(original.host, unpickled.host)
        self.assertEqual(original.user, unpickled.user)
        self.assertEqual(original.msg, unpickled.msg)



from subprocess import TimeoutExpired
from crmsh.report import collect, constants
import crmsh.log

import unittest
from unittest import mock


class TestCollect(unittest.TestCase):

    @mock.patch('logging.Logger.warning')
    @mock.patch('os.path.isfile')
    def test_get_pcmk_log_no_config(self, mock_isfile, mock_warning):
        mock_isfile.side_effect = [False, False, False]
        res = collect.get_pcmk_log()
        self.assertIsNone(res)
        mock_isfile.assert_has_calls([
            mock.call(constants.PCMKCONF),
            mock.call("/var/log/pacemaker/pacemaker.log"),
            mock.call("/var/log/pacemaker.log")
            ])
        mock_warning.assert_called_once_with("No valid pacemaker log file found")

    @mock.patch('logging.Logger.warning')
    @mock.patch('crmsh.utils.read_from_file')
    @mock.patch('os.path.isfile')
    def test_get_pcmk_log(self, mock_isfile, mock_read, mock_warning):
        mock_isfile.return_value = True
        mock_read.return_value = """
# has been enabled, those as well). This log is of more use to developers and
# advanced system administrators, and when reporting problems.
PCMK_logfile=/var/log/pacemaker/pacemaker.log

# Set the permissions on the above log file to owner/group read/write
        """
        res = collect.get_pcmk_log()
        self.assertEqual(res, "/var/log/pacemaker/pacemaker.log")
        mock_isfile.assert_has_calls([
            mock.call(constants.PCMKCONF),
            mock.call("/var/log/pacemaker/pacemaker.log")
            ])
        mock_read.assert_called_once_with(constants.PCMKCONF)

    @mock.patch('crmsh.report.utils.mark_duplicate_basenames')
    @mock.patch('crmsh.report.utils.dump_logset')
    @mock.patch('os.path.isfile')
    @mock.patch('crmsh.report.collect.get_pcmk_log')
    @mock.patch('crmsh.report.collect.get_corosync_log')
    def test_collect_ha_logs(self, mock_corosync_log, mock_get_log, mock_isfile, mock_dump, mock_mark):
        mock_corosync_log.return_value = "/var/log/cluster/corosync.log"
        mock_get_log.return_value = "/var/pacemaker.log"
        mock_mark.return_value = [
            (mock_get_log.return_value, False),
            (mock_corosync_log.return_value, False)
        ]
        mock_isfile.side_effect = [True, True]
        mock_ctx_inst = mock.Mock(extra_log_list=[])

        collect.collect_ha_logs(mock_ctx_inst)

        mock_get_log.assert_called_once_with()
        mock_isfile.assert_has_calls([
            mock.call(mock_get_log.return_value),
            mock.call(mock_corosync_log.return_value)
            ])
        mock_dump.assert_has_calls([
            mock.call(mock_ctx_inst, mock_get_log.return_value, create_dir=False),
            mock.call(mock_ctx_inst, mock_corosync_log.return_value, create_dir=False)
            ])

    @mock.patch('logging.Logger.warning')
    @mock.patch('os.path.exists')
    @mock.patch('crmsh.corosync.conf')
    def test_get_corosync_log_not_exist(self, mock_conf, mock_exists, mock_warning):
        mock_conf.return_value = "/etc/corosync/corosync.conf"
        mock_exists.return_value = False
        self.assertIsNone(collect.get_corosync_log())

    @mock.patch('crmsh.corosync.get_value')
    @mock.patch('os.path.exists')
    @mock.patch('crmsh.corosync.conf')
    def test_get_corosync_log(self, mock_conf, mock_exists, mock_get_value):
        mock_conf.return_value = "/etc/corosync/corosync.conf"
        mock_get_value.return_value = "/var/log/cluster/corosync.log"
        mock_exists.return_value = True
        self.assertEqual(collect.get_corosync_log(), mock_get_value.return_value)

    @mock.patch('crmsh.report.utils.real_path')
    @mock.patch('crmsh.report.collect.logger', spec=crmsh.log.DEBUG2Logger)
    @mock.patch('crmsh.utils.str2file')
    @mock.patch('crmsh.report.utils.get_cmd_output')
    @mock.patch('crmsh.report.utils.ts_to_str')
    def test_collect_journal_logs(self, mock_ts_to_str, mock_get_cmd_output,
                                  mock_str2file, mock_logger, mock_real_path):
        mock_real_path.side_effect = [
                constants.JOURNAL_F,
                constants.JOURNAL_PCMK_F,
                constants.JOURNAL_COROSYNC_F,
                constants.JOURNAL_SBD_F
        ]
        mock_ctx_inst = mock.Mock(from_time=1234, to_time=5678, work_dir="/opt/work")
        mock_ts_to_str.side_effect = ["10.10", "10.12"]
        mock_get_cmd_output.side_effect = ["data_default", "data_pacemaker", "data_corosync", "data_sbd"]
        collect.collect_journal_logs(mock_ctx_inst)
        mock_ts_to_str.assert_has_calls([
            mock.call(mock_ctx_inst.from_time),
            mock.call(mock_ctx_inst.to_time)
            ])
        cmd_list = [
                'journalctl -o short-iso-precise --since "10.10" --until "10.12" --no-pager | tail -n +2',
                'journalctl -u pacemaker -o short-iso-precise --since "10.10" --until "10.12" --no-pager | tail -n +2',
                'journalctl -u corosync -o short-iso-precise --since "10.10" --until "10.12" --no-pager | tail -n +2',
                'journalctl -u sbd -o short-iso-precise --since "10.10" --until "10.12" --no-pager | tail -n +2'
                ]
        mock_get_cmd_output.assert_has_calls([
            mock.call(cmd_list[0]),
            mock.call(cmd_list[1]),
            mock.call(cmd_list[2]),
            mock.call(cmd_list[3]),
            ])
        mock_logger.debug2.assert_has_calls([
            mock.call("Collect journal logs since: 10.10 until: 10.12"),
            mock.call(f"Running command: {cmd_list[0]}"),
            mock.call(f"Running command: {cmd_list[1]}"),
            mock.call(f"Running command: {cmd_list[2]}"),
            mock.call(f"Running command: {cmd_list[3]}"),
            ])
        mock_logger.debug.assert_has_calls([
            mock.call(f"Dump jounal log for default into {constants.JOURNAL_F}"),
            mock.call(f"Dump jounal log for pacemaker into {constants.JOURNAL_PCMK_F}"),
            mock.call(f"Dump jounal log for corosync into {constants.JOURNAL_COROSYNC_F}"),
            mock.call(f"Dump jounal log for sbd into {constants.JOURNAL_SBD_F}")
            ])

    @mock.patch('logging.Logger.debug')
    @mock.patch('os.path.exists')
    def test_collect_sbd_info_no_config(self, mock_exists, mock_debug):
        mock_exists.return_value = False
        mock_ctx_inst = mock.Mock()
        collect.collect_sbd_info(mock_ctx_inst)
        mock_exists.assert_called_once_with(constants.SBDCONF)
        mock_debug.assert_called_once_with(f"SBD config file {constants.SBDCONF} does not exist")

    @mock.patch('shutil.which')
    @mock.patch('shutil.copy2')
    @mock.patch('os.path.exists')
    def test_collect_sbd_info_no_cmd(self, mock_exists, mock_copy, mock_which):
        mock_exists.return_value = True
        mock_which.return_value = False
        mock_ctx_inst = mock.Mock(work_dir="/opt")
        collect.collect_sbd_info(mock_ctx_inst)
        mock_exists.assert_called_once_with(constants.SBDCONF)
        mock_copy.assert_called_once_with(constants.SBDCONF, mock_ctx_inst.work_dir)
        mock_which.assert_called_once_with("sbd")

    @mock.patch('crmsh.report.utils.real_path')
    @mock.patch('builtins.open', create=True)
    @mock.patch('logging.Logger.debug')
    @mock.patch('crmsh.report.utils.get_cmd_output')
    @mock.patch('shutil.which')
    @mock.patch('shutil.copy2')
    @mock.patch('os.path.exists')
    def test_collect_sbd_info(self, mock_exists, mock_copy, mock_which, mock_run, mock_debug, mock_open_file, mock_real_path):
        mock_real_path.return_value = constants.SBD_F
        mock_exists.return_value = True
        mock_which.return_value = True
        mock_open_write = mock.mock_open()
        file_handle = mock_open_write.return_value.__enter__.return_value
        mock_open_file.return_value = mock_open_write.return_value
        mock_run.side_effect = ["data", "data", "data"]
        mock_ctx_inst = mock.Mock(work_dir="/opt")

        collect.collect_sbd_info(mock_ctx_inst)

        mock_exists.assert_called_once_with(constants.SBDCONF)
        mock_copy.assert_called_once_with(constants.SBDCONF, mock_ctx_inst.work_dir)
        mock_which.assert_called_once_with("sbd")
        mock_open_file.assert_called_once_with(f"{mock_ctx_inst.work_dir}/{constants.SBD_F}", "w")
        file_handle.write.assert_has_calls([
            mock.call("\n\n#=====[ Command ] ==========================#\n"),
            mock.call("# . /etc/sysconfig/sbd;export SBD_DEVICE;sbd dump;sbd list\n"),
            mock.call("data"),
            mock.call("\n\n#=====[ Command ] ==========================#\n"),
            mock.call("# crm sbd configure show\n"),
            mock.call("data"),
            mock.call("\n\n#=====[ Command ] ==========================#\n"),
            mock.call("# crm sbd status\n"),
            mock.call("data")
            ])
        mock_debug.assert_called_once_with(f"Dump SBD config file into {constants.SBD_F}")

    @mock.patch('logging.Logger.warning')
    @mock.patch('crmsh.report.collect.ShellUtils')
    def test_pe_to_dot(self, mock_run, mock_warning):
        mock_run_inst = mock.Mock()
        mock_run.return_value = mock_run_inst
        mock_run_inst.get_stdout_stderr.return_value = (1, None, None)
        collect.pe_to_dot("/opt/pe-input-0.bz2")
        mock_run_inst.get_stdout_stderr.assert_called_once_with("crm_simulate -D /opt/pe-input-0.dot -x /opt/pe-input-0.bz2")
        mock_warning.assert_called_once_with('pe_to_dot: %s -> %s failed', '/opt/pe-input-0.bz2', '/opt/pe-input-0.dot')

    @mock.patch('crmsh.report.utils.find_files_in_timespan')
    @mock.patch('crmsh.report.collect.logger', spec=crmsh.log.DEBUG2Logger)
    def test_collect_pe_inputs_no_found(self, mock_logger, mock_find_files):
        mock_ctx_inst = mock.Mock(pe_dir="/opt/pe_dir")
        mock_find_files.return_value = []
        collect.collect_pe_inputs(mock_ctx_inst)
        mock_find_files.assert_called_once_with(mock_ctx_inst, [mock_ctx_inst.pe_dir])
        mock_logger.debug2.assert_has_calls([
            mock.call(f"Looking for PE files in {mock_ctx_inst.pe_dir}"),
            mock.call("No PE file found for the giving time")
            ])

    @mock.patch('crmsh.report.utils.real_path')
    @mock.patch('crmsh.report.collect.pe_to_dot')
    @mock.patch('os.symlink')
    @mock.patch('crmsh.utils.mkdirp')
    @mock.patch('crmsh.report.utils.find_files_in_timespan')
    @mock.patch('crmsh.report.collect.logger', spec=crmsh.log.DEBUG2Logger)
    def test_collect_pe_inputs(self, mock_logger, mock_find_files, mock_mkdir, mock_symlink, mock_to_dot, mock_real_path):
        mock_real_path.return_value = "pe_dir"
        mock_ctx_inst = mock.Mock(pe_dir="/opt/pe_dir", work_dir="/opt/work_dir", speed_up=False)
        mock_find_files.return_value = ["/opt/pe_dir/pe_input1", "/opt/pe_dir/pe_input2"]

        collect.collect_pe_inputs(mock_ctx_inst)

        mock_find_files.assert_called_once_with(mock_ctx_inst, [mock_ctx_inst.pe_dir])
        mock_logger.debug2.assert_has_calls([
            mock.call(f"Looking for PE files in {mock_ctx_inst.pe_dir}"),
            mock.call(f"Found 2 PE files in {mock_ctx_inst.pe_dir}"),
            ])
        mock_logger.debug.assert_called_once_with(f"Dump PE files into pe_dir")

    @mock.patch('crmsh.report.utils.real_path')
    @mock.patch('crmsh.report.collect.logger', spec=crmsh.log.DEBUG2Logger)
    @mock.patch('crmsh.utils.str2file')
    @mock.patch('crmsh.report.utils.get_cmd_output')
    def test_collect_sys_stats(self, mock_run, mock_str2file, mock_logger, mock_real_path):
        mock_real_path.return_value = constants.SYSSTATS_F
        mock_run.side_effect = [
                "data_hostname", "data_uptime", "data_ps_axf", "data_ps_auxw",
                "data_top", "data_ip_addr", "data_ip_link", "data_ip_show", "data_iscsi",
                "data_lspci", "data_mount", "data_cpuinfo", TimeoutExpired("df", 5)
                ]
        mock_ctx_inst = mock.Mock(work_dir="/opt")
        collect.collect_sys_stats(mock_ctx_inst)
        mock_logger.warning.assert_called_once_with(f"Timeout while running command: df")
        mock_run.assert_has_calls([
            mock.call("hostname", timeout=5),
            mock.call("uptime", timeout=5),
            mock.call("ps axf", timeout=5),
            mock.call("ps auxw", timeout=5),
            mock.call("top -b -n 1", timeout=5),
            mock.call("ip addr", timeout=5),
            mock.call("ip -s link", timeout=5),
            mock.call("ip n show", timeout=5),
            mock.call("lsscsi", timeout=5),
            mock.call("lspci", timeout=5),
            mock.call("mount", timeout=5),
            mock.call("cat /proc/cpuinfo", timeout=5),
            mock.call("df", timeout=5)
            ])

    @mock.patch('crmsh.report.utils.real_path')
    @mock.patch('logging.Logger.debug')
    @mock.patch('crmsh.report.utils.get_distro_info')
    @mock.patch('crmsh.utils.str2file')
    @mock.patch('os.uname')
    @mock.patch('crmsh.report.utils.Package')
    def test_collect_sys_info(self, mock_package, mock_uname, mock_str2file, mock_get_distro, mock_debug, mock_real_path):
        mock_real_path.return_value = constants.SYSINFO_F
        mock_package_inst = mock.Mock()
        mock_package.return_value = mock_package_inst
        mock_package_inst.version = mock.Mock(return_value="version_data\n")
        mock_package_inst.verify = mock.Mock(return_value="verify_data\n")
        mock_ctx_inst = mock.Mock(speed_up=False, work_dir="/opt/work")
        mock_uname.return_value = ("Linux", None, "4.5", None, "x86_64")
        mock_get_distro.return_value = "suse"

        collect.collect_sys_info(mock_ctx_inst)

        mock_package.assert_called_once_with(constants.PACKAGES)
        mock_str2file.assert_called_once_with('##### System info #####\nPlatform: Linux\nKernel release: 4.5\nArchitecture: x86_64\nDistribution: suse\n\n##### Installed cluster related packages #####\nversion_data\n\n\n##### Verification output of packages #####\nverify_data\n', '/opt/work/sysinfo.txt')
        mock_debug.assert_called_once_with(f"Dump packages and platform info into {constants.SYSINFO_F}")

    @mock.patch('crmsh.report.utils.real_path')
    @mock.patch('crmsh.report.collect.dump_configurations')
    @mock.patch('crmsh.report.collect.consume_cib_in_workdir')
    @mock.patch('crmsh.report.collect.logger', spec=crmsh.log.DEBUG2Logger)
    @mock.patch('crmsh.utils.str2file')
    @mock.patch('crmsh.report.collect.dump_runtime_state')
    @mock.patch('crmsh.report.collect.ServiceManager')
    def test_collect_config_running(self, mock_service, mock_dump_state, mock_write, mock_debug2, mock_cib, mock_dump_config, mock_real_path):
        mock_real_path.return_value = "workdir"
        mock_service_inst = mock.Mock()
        mock_service.return_value = mock_service_inst
        mock_service_inst.service_is_active.return_value = True
        mock_ctx_inst = mock.Mock(work_dir="/opt/workdir")
        collect.collect_config(mock_ctx_inst)

    @mock.patch('crmsh.report.utils.real_path')
    @mock.patch('crmsh.report.collect.dump_configurations')
    @mock.patch('crmsh.report.collect.consume_cib_in_workdir')
    @mock.patch('crmsh.report.collect.logger', spec=crmsh.log.DEBUG2Logger)
    @mock.patch('crmsh.utils.str2file')
    @mock.patch('shutil.copy2')
    @mock.patch('crmsh.report.collect.ServiceManager')
    def test_collect_config_stopped(self, mock_service, mock_copy2, mock_write, mock_debug2, mock_cib, mock_dump_config, mock_real_path):
        mock_real_path.return_value = "workdir"
        mock_service_inst = mock.Mock()
        mock_service.return_value = mock_service_inst
        mock_service_inst.service_is_active.return_value = False
        mock_ctx_inst = mock.Mock(work_dir="/opt/workdir", cib_dir="/var/log/pacemaker/cib")
        collect.collect_config(mock_ctx_inst)

    @mock.patch('crmsh.utils.str2file')
    @mock.patch('crmsh.report.collect.sh.cluster_shell')
    @mock.patch('os.path.isfile')
    def test_consume_cib_in_workdir(self, mock_isfile, mock_run, mock_str2file):
        mock_isfile.return_value = True
        mock_run_inst = mock.Mock()
        mock_run.return_value = mock_run_inst
        mock_run_inst.get_stdout_or_raise_error.return_value = "data1"
        mock_run_inst.get_rc_stdout_stderr_without_input.return_value = (0, "data2", "error")
        collect.consume_cib_in_workdir("/workdir")
        mock_isfile.assert_called_once_with(f"/workdir/{constants.CIB_F}")
        cmd1 = "CIB_file=/workdir/cib.xml crm configure show"
        mock_run_inst.get_stdout_or_raise_error.assert_called_once_with(cmd1)
        cmd2 = f"crm_verify -V -x /workdir/cib.xml"
        mock_run_inst.get_rc_stdout_stderr_without_input.assert_called_once_with(None, cmd2)
        mock_str2file.assert_has_calls([
            mock.call("data1", f"/workdir/{constants.CONFIGURE_SHOW_F}"),
            mock.call("error", f"/workdir/{constants.CRM_VERIFY_F}")
        ])

    @mock.patch('crmsh.report.utils.real_path')
    @mock.patch('crmsh.report.collect.logger', spec=crmsh.log.DEBUG2Logger)
    @mock.patch('shutil.copy2')
    @mock.patch('crmsh.utils.mkdirp')
    @mock.patch('crmsh.report.utils.find_files_in_timespan')
    def test_collect_ratraces(self, mock_find, mock_mkdirp, mock_copy, mock_logger, mock_real_path):
        mock_real_path.return_value = "/var/log"
        data = "INFO: Trace for .* is written to /var/log/cluster/pacemaker.log"
        mock_ctx_inst = mock.Mock(
            work_dir="/opt/work",
            trace_dir_list="/var/lib/heartbeat/trace_ra",
        )
        mock_find.return_value = [
            "/var/lib/heartbeat/trace_ra/IPaddr2/admin-ip.monitor.2024-02-26.15:21:39",
            "/var/lib/heartbeat/trace_ra/IPaddr2/admin-ip.start.2024-02-26.15:21:39",
            "/var/lib/heartbeat/trace_ra/IPaddr2/admin-ip.stop.2024-02-26.15:21:46",
        ]

        collect.collect_ratraces(mock_ctx_inst)

        mock_mkdirp.assert_called_with('/opt/work/trace_ra/IPaddr2')
        mock_copy.assert_has_calls([
            mock.call("/var/lib/heartbeat/trace_ra/IPaddr2/admin-ip.monitor.2024-02-26.15:21:39", '/opt/work/trace_ra/IPaddr2'),
            mock.call("/var/lib/heartbeat/trace_ra/IPaddr2/admin-ip.start.2024-02-26.15:21:39", '/opt/work/trace_ra/IPaddr2'),
            mock.call("/var/lib/heartbeat/trace_ra/IPaddr2/admin-ip.stop.2024-02-26.15:21:46", '/opt/work/trace_ra/IPaddr2'),
        ])
        mock_logger.debug.assert_called_with(f'Dump RA trace files into {mock_real_path.return_value}')

    @mock.patch('crmsh.report.utils.real_path')
    @mock.patch('logging.Logger.debug')
    @mock.patch('crmsh.utils.str2file')
    @mock.patch('crmsh.report.utils.get_cmd_output')
    @mock.patch('shutil.which')
    def test_collect_dlm_info(self, mock_which, mock_get_output, mock_str2file, mock_debug, mock_real_path):
        mock_real_path.return_value = constants.DLM_DUMP_F
        mock_which.return_value = True
        ls_data = """
dlm lockspaces
name          08BB5A6A38EE491DBF63627EEB57E558
id            0x19041a12
        """
        mock_get_output.side_effect = [ls_data, "lockdebug data", "dump data"]
        mock_ctx_inst = mock.Mock(work_dir="/opt/work_dir")
        collect.collect_dlm_info(mock_ctx_inst)
        mock_debug.assert_called_once_with(f"Dump DLM information into {constants.DLM_DUMP_F}")

    @mock.patch('crmsh.report.collect.dump_core_info')
    @mock.patch('logging.Logger.warning')
    @mock.patch('os.path.basename')
    @mock.patch('crmsh.report.utils.find_files_in_timespan')
    def test_collect_coredump_info(self, mock_find, mock_basename, mock_warning, mock_dump):
        mock_ctx_inst = mock.Mock(cores_dir_list=['/var/lib/pacemaker/cores'], work_dir="/opt/work_dir")
        mock_find.return_value = ["/var/lib/pacemaker/cores/core.1"]
        mock_basename.return_value = "core.1"
        collect.collect_coredump_info(mock_ctx_inst)
        mock_dump.assert_called_once_with("/opt/work_dir", mock_find.return_value)
        mock_warning.assert_called_once_with(f"Found coredump file: {mock_find.return_value}")

    @mock.patch('crmsh.report.collect.ShellUtils')
    def test_find_binary_path_for_core_not_found(self, mock_run):
        mock_run().get_stdout_stderr.return_value = (0, "Core not found", None)
        res = collect.find_binary_path_for_core("core.1")
        self.assertEqual("Cannot find the program path for core core.1", res)

    @mock.patch('crmsh.report.collect.ShellUtils')
    def test_find_binary_path_for_core(self, mock_run):
        mock_run_inst = mock.Mock()
        mock_run.return_value = mock_run_inst
        mock_run_inst.get_stdout_stderr.return_value = (0, "Core was generated by `/usr/sbin/crm_mon'", None)
        res = collect.find_binary_path_for_core("core.1")
        self.assertEqual("Core core.1 was generated by /usr/sbin/crm_mon", res)

    @mock.patch('crmsh.report.utils.real_path')
    @mock.patch('crmsh.report.collect.logger', spec=crmsh.log.DEBUG2Logger)
    @mock.patch('crmsh.utils.str2file')
    @mock.patch('shutil.which')
    def test_dump_core_info_no_gdb(self, mock_which, mock_str2file, mock_logger, mock_real_path):
        mock_real_path.return_value = constants.COREDUMP_F
        mock_which.return_value = False
        collect.dump_core_info("/opt/workdir", ["core.1"])
        mock_logger.warning.assert_called_once_with("Please install gdb to get more info for coredump files")

    @mock.patch('crmsh.report.utils.real_path')
    @mock.patch('crmsh.report.collect.logger', spec=crmsh.log.DEBUG2Logger)
    @mock.patch('crmsh.utils.str2file')
    @mock.patch('crmsh.report.collect.find_binary_path_for_core')
    @mock.patch('shutil.which')
    def test_dump_core_info(self, mock_which, mock_find_binary, mock_str2file, mock_debug2, mock_real_path):
        mock_real_path.return_value = constants.COREDUMP_F
        mock_which.return_value = True
        mock_find_binary.return_value = "data"
        collect.dump_core_info("/opt/workdir", ["core.1"])
        mock_str2file.assert_called_once_with("data\n\nPlease utilize the gdb and debuginfo packages to obtain more detailed information locally", f"/opt/workdir/{constants.COREDUMP_F}")
        mock_debug2(f"Dump coredump info into {constants.COREDUMP_F}")

    @mock.patch('crmsh.utils.str2file')
    @mock.patch('pwd.getpwnam')
    @mock.patch('os.stat')
    @mock.patch('os.path.isdir')
    def test_collect_perms_state(self, mock_isdir, mock_stat, mock_getpwnam, mock_str2file):
        mock_ctx_inst = mock.Mock(
            pcmk_lib_dir="/var/lib/pacemaker",
            pe_dir="/var/lib/pacemaker/pe",
            cib_dir="/var/lib/pacemaker/cib",
            work_dir="/opt/work_dir"
        )
        mock_isdir.side_effect = [False, True, True]
        mock_stat_inst_pe = mock.Mock(st_uid=1000, st_gid=1000, st_mode=0o750)
        mock_stat_inst_cib = mock.Mock(st_uid=1000, st_gid=1000, st_mode=0o750)
        mock_stat.side_effect = [mock_stat_inst_pe, mock_stat_inst_cib]
        mock_getpwnam_inst_pe = mock.Mock(pw_uid=1000, pw_gid=1000)
        mock_getpwnam_inst_cib = mock.Mock(pw_uid=1001, pw_gid=1000)
        mock_getpwnam.side_effect = [mock_getpwnam_inst_pe, mock_getpwnam_inst_cib]

        collect.collect_perms_state(mock_ctx_inst)

        data = "##### Check perms for /var/lib/pacemaker: /var/lib/pacemaker is not a directory or does not exist\n##### Check perms for /var/lib/pacemaker/pe: OK\n##### Check perms for /var/lib/pacemaker/cib: Permissions or ownership for /var/lib/pacemaker/cib are incorrect\n"
        mock_str2file.assert_called_once_with(data, f"/opt/work_dir/{constants.PERMISSIONS_F}")

    @mock.patch('crmsh.report.utils.real_path')
    @mock.patch('crmsh.utils.this_node')
    @mock.patch('crmsh.utils.get_dc')
    @mock.patch('logging.Logger.debug')
    @mock.patch('crmsh.utils.str2file')
    @mock.patch('crmsh.report.collect.sh.cluster_shell')
    def test_dump_runtime_state(self, mock_run, mock_str2file, mock_debug, mock_get_dc, mock_this_node, mock_real_path):
        mock_real_path.side_effect = [
            constants.CRM_MON_F,
            constants.CIB_F,
            constants.MEMBERSHIP_F,
            "workdir"
        ]
        mock_run_inst = mock.Mock()
        mock_run.return_value = mock_run_inst
        mock_run_inst.get_stdout_or_raise_error.side_effect = [
                "crm_mon_data_r1",
                "crm_mon_data_n1",
                "crm_mon_data_rf1",
                "crm_mon_data_rnt1",
                "cib_data",
                "crm_node_data"
                ]
        mock_get_dc.return_value = "node1"
        mock_this_node.return_value = "node1"
        collect.dump_runtime_state("/opt/workdir")
        mock_debug.assert_has_calls([
            mock.call(f"Dump crm_mon state into {constants.CRM_MON_F}"),
            mock.call(f"Dump CIB contents into {constants.CIB_F}"),
            mock.call(f"Dump members of this partition into {constants.MEMBERSHIP_F}"),
            mock.call(f"Current DC is node1; Touch file 'DC' in workdir")
        ])

    @mock.patch('shutil.copytree')
    @mock.patch('os.path.basename')
    @mock.patch('os.path.isdir')
    @mock.patch('shutil.copy2')
    @mock.patch('os.path.isfile')
    @mock.patch('crmsh.corosync.conf')
    def test_dump_configurations(self, mock_corosync_conf, mock_isfile, mock_copy2, mock_isdir, mock_basename, mock_copytree):
        mock_corosync_conf.return_value = "/etc/corosync/corosync.conf"
        mock_isfile.side_effect = [True, True, False, True]
        mock_isdir.return_value = True
        mock_basename.return_value = "drbd.d"
        collect.dump_configurations("/opt/workdir")

    @mock.patch('crmsh.report.utils.real_path')
    @mock.patch('logging.Logger.debug')
    @mock.patch('crmsh.utils.str2file')
    @mock.patch('crmsh.report.utils.get_cmd_output')
    @mock.patch('crmsh.report.utils.find_files_in_timespan')
    def test_collect_corosync_blackbox(self, mock_find_files, mock_get_cmd_output, mock_str2file, mock_debug, mock_real_path):
        mock_real_path.return_value = constants.COROSYNC_RECORDER_F
        mock_ctx_inst = mock.Mock(work_dir="/opt/workdir")
        mock_find_files.return_value = ["/var/lib/corosync/fdata.1"]
        mock_get_cmd_output.return_value = "data"
        collect.collect_corosync_blackbox(mock_ctx_inst)
        mock_debug.assert_called_once_with(f"Dump corosync blackbox info into {constants.COROSYNC_RECORDER_F}")

    @mock.patch('crmsh.corosync.query_quorum_status')
    @mock.patch('crmsh.report.collect.ServiceManager')
    def test_collect_qdevice_info_return(self, mock_service, mock_quorum):
        mock_ctx_inst = mock.Mock(work_dir="/opt/workdir")
        mock_service_inst = mock.Mock()
        mock_service.return_value = mock_service_inst
        mock_service_inst.service_is_active.return_value = False
        collect.collect_qdevice_info(mock_ctx_inst)
        mock_quorum.assert_not_called()

    @mock.patch('logging.Logger.debug')
    @mock.patch('crmsh.report.utils.real_path')
    @mock.patch('crmsh.utils.str2file')
    @mock.patch('os.path.join')
    @mock.patch('crmsh.corosync.query_qnetd_status')
    @mock.patch('crmsh.corosync.query_qdevice_status')
    @mock.patch('crmsh.corosync.query_quorum_status')
    @mock.patch('crmsh.report.collect.ServiceManager')
    def test_collect_qdevice_info(self, mock_service, mock_quorum, mock_qdevice, mock_qnetd, mock_join, mock_str2file, mock_real_path, mock_debug):
        mock_ctx_inst = mock.Mock(work_dir="/opt/workdir")
        mock_service_inst = mock.Mock()
        mock_service.return_value = mock_service_inst
        mock_service_inst.service_is_active.side_effect = [True, True]
        mock_qdevice.return_value = "qdevice_data"
        mock_qnetd.return_value = "qnetd_data"
        mock_quorum.return_value = "quorum_data"
        mock_join.return_value = "/opt/workdir/qdevice.txt"
        mock_real_path.return_value = "/opt/workdir/qdevice.txt"
        collect.collect_qdevice_info(mock_ctx_inst)
        mock_debug.assert_called_once_with('Dump quorum/qdevice/qnetd information into /opt/workdir/qdevice.txt')

    @mock.patch("logging.Logger.error")
    @mock.patch('crmsh.utils.str2file')
    @mock.patch("crmsh.report.utils.real_path")
    @mock.patch("logging.Logger.debug")
    @mock.patch("crmsh.report.collect.cluster_fs_commands_output")
    @mock.patch("crmsh.report.collect.lsof_cluster_fs_device")
    @mock.patch("crmsh.report.collect.dump_D_process")
    @mock.patch("crmsh.report.collect.ShellUtils")
    @mock.patch("shutil.which")
    def test_collect_cluster_fs_info(self, mock_which, mock_run, mock_dump, mock_lsof, mock_cluster, mock_debug, mock_real_path, mock_str2file, mock_error):
        mock_which.return_value = True
        mock_run_inst = mock.Mock()
        mock_run.return_value = mock_run_inst
        mock_run_inst.get_stdout_stderr.side_effect = [
            (1, None, "error"),
            (0, "data", None)
        ]
        mock_dump.return_value = "dump_data\n"
        mock_lsof.return_value = "lsof_data\n"
        mock_cluster.return_value = "cluster_data"
        mock_ctx_inst = mock.Mock(work_dir="/opt/work")
        mock_real_path.side_effect = ["/path/ocfs2.txt", "/path/gfs2.txt"]

        collect.collect_cluster_fs_info(mock_ctx_inst)

        mock_debug.assert_has_calls([
            mock.call('Dump %s information into %s', 'OCFS2', '/path/ocfs2.txt'),
            mock.call('Dump %s information into %s', 'GFS2', '/path/gfs2.txt')
        ])
        mock_str2file.assert_has_calls([
            mock.call('Failed to run "mounted.ocfs2 -d": error', '/opt/work/ocfs2.txt'),
            mock.call('dump_data\nlsof_data\ncluster_data', '/opt/work/gfs2.txt')
        ])

    @mock.patch('crmsh.report.collect.ShellUtils')
    def test_dump_D_process_empty(self, mock_run):
        mock_run_inst = mock.Mock()
        mock_run.return_value = mock_run_inst
        mock_run_inst.get_stdout_stderr.return_value = (0, None, None)
        res = collect.dump_D_process()
        self.assertEqual(res, "Dump D-state process stack: 0\n")

    @mock.patch('crmsh.report.collect.ShellUtils')
    def test_dump_D_process(self, mock_run):
        mock_run_inst = mock.Mock()
        mock_run.return_value = mock_run_inst
        mock_run_inst.get_stdout_stderr.side_effect = [
            (0, "1000", None),
            (0, "data1", None),
            (0, "data2", None)
        ]
        res = collect.dump_D_process()
        self.assertEqual(res, "Dump D-state process stack: 1\npid: 1000     comm: data1\ndata2\n\n")
        mock_run_inst.get_stdout_stderr.assert_has_calls([
            mock.call("ps aux|awk '$8 ~ /^D/{print $2}'"),
            mock.call('cat /proc/1000/comm'),
            mock.call('cat /proc/1000/stack')
            ])

    @mock.patch('crmsh.report.collect.ShellUtils')
    def test_lsof_cluster_fs_device(self, mock_run):
        mock_run_inst = mock.Mock()
        mock_run.return_value = mock_run_inst
        mount_data = """
/dev/vda3 on /home type xfs (rw,relatime,attr2,inode64,logbufs=8,logbsize=32k,noquota)
tmpfs on /run/user/0 type tmpfs (rw,nosuid,nodev,relatime,size=169544k,nr_inodes=42386,mode=700,inode64)
/dev/sda7 on /srv/clusterfs type ocfs2 (rw,relatime,heartbeat=non
        """
        mock_run_inst.get_stdout_stderr.side_effect = [(0, mount_data, None), (0, "data", None)]
        res = collect.lsof_cluster_fs_device("OCFS2")
        self.assertEqual(res, "\n\n#=====[ Command ] ==========================#\n# lsof /dev/sda7\ndata")
        mock_run_inst.get_stdout_stderr.assert_has_calls([
            mock.call("mount"),
            mock.call("lsof /dev/sda7")
        ])

    @mock.patch('crmsh.report.utils.get_cmd_output')
    @mock.patch('os.path.exists')
    @mock.patch('shutil.which')
    def test_cluster_fs_commands_output(self, mock_which, mock_exists, mock_run):
        mock_which.side_effect = [False for i in range(5)] + [True, True]
        mock_exists.return_value = False
        mock_run.return_value = "data"
        res = collect.cluster_fs_commands_output("OCFS2")
        self.assertEqual(res, "\n\n#===== [ Command ] ==========================#\n# mounted.ocfs2 -f\ndata")



# Copyright (C) 2013-2017 Daniel Fairhead
# Copyright (C) 2017 Kristoffer Gronlund <kgronlund@suse.com>
# See COPYING for license information.

"""
Based on simpleeval:

SimpleEval - (C) 2013-2017 Daniel Fairhead
-------------------------------------

An short, easy to use, safe and reasonably extensible expression evaluator.
Designed for things like in a website where you want to allow the user to
generate a string, or a number from some other input, without allowing full
eval() or other unsafe or needlessly complex linguistics.

-------------------------------------

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

-------------------------------------

Initial idea copied from J.F. Sebastian on Stack Overflow
( http://stackoverflow.com/a/9558001/1973500 ) with
modifications and many improvments.

-------------------------------------
Contributors:
- corro (Robin Baumgartner) (py3k)
- dratchkov (David R) (nested dicts)
- marky1991 (Mark Young) (slicing)
- T045T (Nils Berg) (!=, py3kstr, obj.
- perkinslr (Logan Perkins) (.__globals__ or .func_ breakouts)
- impala2 (Kirill Stepanov) (massive _eval refactor)
- gk (ugik) (Other iterables than str can DOS too, and can be made)
- daveisfera (Dave Johansen) 'not' Boolean op, Pycharm and pep8 fixes.

-------------------------------------
Usage:

>>> s = SimpleEval()
>>> s.eval("20 + 30")
50

You can add your own functions easily too:

if file.txt contents is "11"

>>> def get_file():
        with open("file.txt",'r') as f:
            return f.read()

    s.functions["get_file"] = get_file
    s.eval("int(get_file()) + 31")
42

For more information, see the full package documentation on pypi, or the github
repo.

-----------

If you don't need to re-use the evaluator (with it's names, functions, etc),
then you can use the simple_eval() function:

>>> simple_eval("21 + 19")
40

You can pass names, operators and functions to the simple_eval function as
well:

>>> simple_eval("40 + two", names={"two": 2})
42

"""

import ast
import sys
import operator as op

########################################
# Module wide 'globals'

MAX_STRING_LENGTH = 100000
DISALLOW_PREFIXES = ['_', 'func_']
DISALLOW_METHODS = ['format']

PYTHON3 = sys.version_info[0] == 3

########################################
# Exceptions:


class InvalidExpression(Exception):
    """ Generic Exception """
    pass


class NameNotDefined(InvalidExpression):
    """ a name isn't defined. """
    def __init__(self, name, expression):
        self.name = name
        self.message = "'{0}' is not defined for expression '{1}'".format(
            name, expression)
        self.expression = expression

        # pylint: disable=bad-super-call
        super(InvalidExpression, self).__init__(self.message)


class AttributeDoesNotExist(InvalidExpression):
    """attribute does not exist"""
    def __init__(self, attr, expression):
        self.message = \
            "Attribute '{0}' does not exist in expression '{1}'".format(
                attr, expression)
        self.attr = attr
        self.expression = expression


class FeatureNotAvailable(InvalidExpression):
    """ What you're trying to do is not allowed. """
    pass


class IterableTooLong(InvalidExpression):
    """ That iterable is **way** too long, baby. """
    pass


########################################
# Defaults for the evaluator:

DEFAULT_OPERATORS = {ast.Eq: op.eq, ast.NotEq: op.ne,
                     ast.Gt: op.gt, ast.Lt: op.lt,
                     ast.GtE: op.ge, ast.LtE: op.le,
                     ast.Not: op.not_,
                     ast.USub: op.neg, ast.UAdd: op.pos,
                     ast.In: lambda x, y: op.contains(y, x),
                     ast.NotIn: lambda x, y: not op.contains(y, x),
                     ast.Is: lambda x, y: x is y,
                     ast.IsNot: lambda x, y: x is not y,
                     }

DEFAULT_NAMES = {"True": True, "False": False}

########################################
# And the actual evaluator:


class SimpleEval(object):  # pylint: disable=too-few-public-methods
    """ A very simple expression parser.
        >>> s = SimpleEval()
        >>> s.eval("20 + 30 - ( 10 * 5)")
        0
        """
    expr = ""

    def __init__(self, names):
        """
            Create the evaluator instance.  Set up valid operators (+,-, etc)
            functions (add, random, get_val, whatever) and names. """

        operators = DEFAULT_OPERATORS
        names = names.copy()
        names.update(DEFAULT_NAMES)

        self.operators = operators
        self.names = names

        self.nodes = {
            ast.Num: self._eval_num,
            ast.Str: self._eval_str,
            ast.Name: self._eval_name,
            ast.UnaryOp: self._eval_unaryop,
            ast.BinOp: self._eval_binop,
            ast.BoolOp: self._eval_boolop,
            ast.Compare: self._eval_compare,
            ast.IfExp: self._eval_ifexp,
            ast.keyword: self._eval_keyword,
            ast.Subscript: self._eval_subscript,
            ast.Attribute: self._eval_attribute,
            ast.Index: self._eval_index,
            ast.Slice: self._eval_slice,
        }

        # py3k stuff:
        if hasattr(ast, 'NameConstant'):
            self.nodes[ast.NameConstant] = self._eval_constant
        elif isinstance(self.names, dict) and "None" not in self.names:
            self.names["None"] = None

        # py3.8 uses ast.Constant instead of ast.Num, ast.Str, ast.NameConstant
        if hasattr(ast, 'Constant'):
            self.nodes[ast.Constant] = self._eval_constant

    def evaluate(self, expr):
        """ evaluate an expresssion, using the operators, functions and
            names previously set up. """

        # set a copy of the expression aside, so we can give nice errors...

        self.expr = expr

        # and evaluate:
        return self._eval(ast.parse(expr.strip()).body[0].value)

    def _eval(self, node):
        """ The internal evaluator used on each node in the parsed tree. """

        try:
            handler = self.nodes[type(node)]
        except KeyError:
            raise FeatureNotAvailable("Sorry, {0} is not available in this "
                                      "evaluator".format(type(node).__name__))

        return handler(node)

    @staticmethod
    def _eval_num(node):
        return node.n

    @staticmethod
    def _eval_str(node):
        if len(node.s) > MAX_STRING_LENGTH:
            raise IterableTooLong("String Literal in statement is too long!"
                                  " ({0}, when {1} is max)".format(
                                      len(node.s), MAX_STRING_LENGTH))
        return node.s

    @staticmethod
    def _eval_constant(node):
        if (hasattr(node.value, '__len__') and
                len(node.value) > MAX_STRING_LENGTH):
            raise IterableTooLong("Literal in statement is too long!"
                                  " ({0}, when {1} is max)"
                                  "".format(len(node.value),
                                            MAX_STRING_LENGTH))
        return node.value

    def _eval_unaryop(self, node):
        return self.operators[type(node.op)](self._eval(node.operand))

    def _eval_binop(self, node):
        return self.operators[type(node.op)](self._eval(node.left),
                                             self._eval(node.right))

    def _eval_boolop(self, node):
        if isinstance(node.op, ast.And):
            vout = False
            for value in node.values:
                vout = self._eval(value)
                if not vout:
                    return False
            return vout
        elif isinstance(node.op, ast.Or):
            for value in node.values:
                vout = self._eval(value)
                if vout:
                    return vout
            return False

    def _eval_compare(self, node):
        left = self._eval(node.left)
        for operation, comp in zip(node.ops, node.comparators):
            right = self._eval(comp)
            if self.operators[type(operation)](left, right):
                left = right  # Hi Dr. Seuss...
            else:
                return False
        return True

    def _eval_ifexp(self, node):
        return self._eval(node.body) if self._eval(node.test) \
                                         else self._eval(node.orelse)

    def _eval_keyword(self, node):
        return node.arg, self._eval(node.value)

    def _eval_name(self, node):
        try:
            # This happens at least for slicing
            # This is a safe thing to do because it is impossible
            # that there is a true exression assigning to none
            # (the compiler rejects it, so you can't even
            # pass that to ast.parse)
            if isinstance(self.names, dict):
                return self.names[node.id]
            elif callable(self.names):
                return self.names(node)
            else:
                raise InvalidExpression('Trying to use name (variable) "{0}"'
                                        ' when no "names" defined for'
                                        ' evaluator'.format(node.id))

        except KeyError:
            if node.id in self.functions:
                return self.functions[node.id]

            raise NameNotDefined(node.id, self.expr)

    def _eval_subscript(self, node):

        container = self._eval(node.value)
        key = self._eval(node.slice)
        try:
            return container[key]
        except KeyError:
            raise

    def _eval_attribute(self, node):
        for prefix in DISALLOW_PREFIXES:
            if node.attr.startswith(prefix):
                raise FeatureNotAvailable(
                    "Sorry, access to __attributes "
                    " or func_ attributes is not available. "
                    "({0})".format(node.attr))
        if node.attr in DISALLOW_METHODS:
            raise FeatureNotAvailable(
                    "Sorry, this method is not available. "
                    "({0})".format(node.attr))

        try:
            return self._eval(node.value)[node.attr]
        except (KeyError, TypeError):
            pass

        # Maybe the base object is an actual object, not just a dict
        try:
            return getattr(self._eval(node.value), node.attr)
        except (AttributeError, TypeError):
            pass

        # If it is neither, raise an exception
        raise AttributeDoesNotExist(node.attr, self.expr)

    def _eval_index(self, node):
        return self._eval(node.value)

    def _eval_slice(self, node):
        lower = upper = step = None
        if node.lower is not None:
            lower = self._eval(node.lower)
        if node.upper is not None:
            upper = self._eval(node.upper)
        if node.step is not None:
            step = self._eval(node.step)
        return slice(lower, upper, step)


def minieval(expr, env):
    """
    Given a dict of variable -> value mapping in env,
    parse and evaluate the expression in expr
    """
    return SimpleEval(env).evaluate(expr)



# Copyright (C) 2017 Xin Liang <XLiang@suse.com>
# See COPYING for license information.
BIN_CRM = "/usr/sbin/crm"
BIN_COLLECTOR = f"{BIN_CRM} report __collector"
COMPRESS_DATA_FLAG = "COMPRESS CRM_REPORT DATA:::"
LOG_PATTERNS = ["CRIT:", "ERROR:", "WARNING:", "crit:", "error:", "warning:"]
PTEST = "crm_simulate"
SSH_OPTS = "-o StrictHostKeyChecking=no -o EscapeChar=none -o ConnectTimeout=15"
CHECK_LOG_LINES = 10
STAMP_TYPE = ""

DECRIPTION_TMPLATE = """
Please edit this template and describe the issue/problem you
encountered. Then, post to
    http://clusterlabs.org/mailman/listinfo/users
or file a bug at
    https://github.com/ClusterLabs/crmsh/issues

Thank you.

Date: {0}
By: crm report {1}
Subject: [short problem description]
Severity: [choose one] enhancement minor normal major critical blocking
--------------------------------------------------------

Detailed description:

"""

PACKAGES = "booth corosync corosync-qdevice corosync-qnetd corosync-testagents crmsh crmsh-scripts csync2 doxygen2man drbd-utils gfs2-kmp-default gfs2-utils hawk-apiserver ldirectord libcfg6 libcmap4 libcorosync_common4 libcpg4 libdlm libdlm3 libqb-tools libqb100 libquorum5 libsam4 libtotem_pg5 libvotequorum8 linstor linstor-common linstor-controller linstor-satellite monitoring-plugins-metadata o2locktop ocfs2-tools ocfs2-tools-o2cb omping pacemaker pacemaker-cli pacemaker-cts pacemaker-libs pacemaker-remote pacemaker-schemas patterns-ha pssh python-pssh python3-linstor python3-linstor-client python3-pacemaker python3-parallax resource-agents resource-agents-zfs ruby2.5-rubygem-sass-listen ruby2.5-rubygem-sass-listen-doc sbd"

ANALYSIS_F = "analysis.txt"
COREDUMP_F = "coredump_info.txt"
CIB_F = "cib.xml"
CONFIGURE_SHOW_F = "configure_show.txt"
CONFIGURATIONS = [
    "/etc/drbd.conf",
    "/etc/drbd.d",
    "/etc/booth/booth.conf"
]
COROSYNC_RECORDER_F = "fdata.txt"
COROSYNC_F = "corosync.conf"
CRM_MON_F = "crm_mon.txt"
CRM_VERIFY_F = "crm_verify.txt"
DESCRIPTION_F = "description.txt"
DLM_DUMP_F = "dlm_dump.txt"
JOURNAL_F = "journal.log"
JOURNAL_PCMK_F = "journal_pacemaker.log"
JOURNAL_COROSYNC_F = "journal_corosync.log"
JOURNAL_SBD_F = "journal_sbd.log"
MEMBERSHIP_F = "members.txt"
PERMISSIONS_F = "permissions.txt"
SBDCONF = "/etc/sysconfig/sbd"
PCMKCONF = "/etc/sysconfig/pacemaker"
SYSINFO_F = "sysinfo.txt"
SYSSTATS_F = "sysstats.txt"
OCFS2_F = "ocfs2.txt"
GFS2_F = "gfs2.txt"
SBD_F = "sbd.txt"
QDEVICE_F = "quorum_qdevice_qnetd.txt"
OSRELEASE = "/etc/os-release"
TIME_FORMAT = "%Y-%m-%d %H:%M:%S"
RESULT_TIME_SUFFIX = "%a-%d-%b-%Y"
NAME = "crm report"
COROSYNC_LIB = "/var/lib/corosync"

DESCRIPTION_HELP = '''Description:
crm report is a utility to collect all information (logs,
configuration files, system information, etc) relevant to
Pacemaker (CRM) over the given period of time.
'''

EXTRA_HELP = '''
Examples
  # collect from 2pm, today
  {name} -f 2pm report_1

  # collect from "2007/9/5 12:30" to "2007/9/5 14:00"
  {name} -f "2007/9/5 12:30" -t "2007/9/5 14:00" report_2

  # collect from 1:00 to 3:00, today; include /var/log/cluster/ha-debug as extra log
  {name} -f 1:00 -t 3:00 -E /var/log/cluster/ha-debug report_3

  # collect from "09sep07 2:00" and use 'hacluster' as ssh user
  {name} -f "09sep07 2:00" -u hacluster report_4

  # collect from 18:00, today; replace sensitive message like "usern.*" or "admin.*"
  {name} -f 18:00 -s -p "usern.*" -p "admin.*" report_5

  # collect from 1 mounth ago
  {name} -f 1m

  # collect from 75 hours ago
  {name} -f 75H

  # collect from 10 minutes ago
  {name} -f 10M

  # collect from 2 days ago to 1 day ago
  {name} -f 2d -t 1d
'''.format(name=NAME)
# vim:ts=4:sw=4:et:



import logging
import os.path
import typing

import sys

import crmsh.healthcheck
import crmsh.parallax
import crmsh.utils
from crmsh import user_of_host
import crmsh.config
import crmsh.constants
from crmsh.prun import prun


# pump this seq when upgrade check need to be run
CURRENT_UPGRADE_SEQ = (1, 0)
DATA_DIR = '/var/lib/crmsh'
SEQ_FILE_PATH = DATA_DIR + '/upgrade_seq'
# touch this file to force a upgrade process
FORCE_UPGRADE_FILE_PATH = DATA_DIR + '/upgrade_forced'

TIMEOUT_SSH_GET_SEQUENCE = 3


VERSION_FEATURES = {
    (1, 0): [crmsh.healthcheck.PasswordlessHaclusterAuthenticationFeature]
}


logger = logging.getLogger(__name__)


class _SkipUpgrade(Exception):
    pass


def _parse_upgrade_seq(s: bytes) -> typing.Tuple[int, int]:
    parts = s.split(b'.', 1)
    if len(parts) != 2:
        raise ValueError('Invalid upgrade seq {}'.format(s))
    major = int(parts[0])
    minor = int(parts[1])
    return major, minor


def _format_upgrade_seq(s: typing.Tuple[int, int]) -> str:
    return '.'.join((str(x) for x in s))


def _get_file_content(path, default=None):
    try:
        with open(path, 'rb') as f:
            return f.read()
    except FileNotFoundError:
        return default


def _parallax_run(nodes: str, cmd: str) -> typing.Dict[str, typing.Tuple[int, bytes, bytes]]:
    results = prun.prun({node: cmd for node in nodes})
    for node, result in results.items():
        if isinstance(result, prun.SSHError):
            raise ValueError("Failed on {}: {}".format(node, result))
    return {node: (result.returncode, result.stdout, result.stderr) for node, result in results.items()}


def _is_upgrade_needed(nodes):
    """decide whether upgrading is needed by checking local sequence file"""
    needed = False
    try:
        os.stat(FORCE_UPGRADE_FILE_PATH)
        needed = True
    except FileNotFoundError:
        pass
    if not needed:
        s = _get_file_content(SEQ_FILE_PATH, b'').strip()
        if s == b'':
            # try the old path
            seq_file_path = os.path.expanduser('~hacluster/crmsh') + '/upgrade_seq'
            s = _get_file_content(seq_file_path, b'').strip()
            if s != b'':
                try:
                    os.mkdir(DATA_DIR)
                except FileExistsError:
                    pass
                with open(SEQ_FILE_PATH, 'wb') as f:
                    f.write(s)
                    f.write(b'\n')
        try:
            local_seq = _parse_upgrade_seq(s)
        except ValueError:
            local_seq = (0, 0)
        needed = CURRENT_UPGRADE_SEQ > local_seq
    return needed


def _is_cluster_target_seq_consistent(nodes):
    cmd = '/usr/bin/env python3 -m crmsh.upgradeutil get-seq'
    try:
        results = prun.prun({node: cmd for node in nodes}, timeout_seconds=TIMEOUT_SSH_GET_SEQUENCE)
        for node, result in results.items():
            if isinstance(result, prun.PRunError):
                logger.debug("upgradeutil: get-seq failed: %s", result)
                raise _SkipUpgrade() from None
    except (prun.PRunError, user_of_host.UserNotFoundError) as e:
        logger.debug("upgradeutil: get-seq failed: %s", e)
        raise _SkipUpgrade() from None
    ret = True
    for node, result in results.items():
        if result.returncode != 0:
            logger.debug(
                "upgradeutil: remote command '%s' fails on %s: rc=%s, stdout=%s, stderr=%s",
                cmd, node, result.returncode, result.stdout, result.stderr,
            )
            raise _SkipUpgrade() from None
        else:
            try:
                seq = _parse_upgrade_seq(result.stdout.strip())
                if CURRENT_UPGRADE_SEQ != seq:
                    ret = False
                    logger.debug('seq %s from %s != %s', seq, node, CURRENT_UPGRADE_SEQ)
            except ValueError as e:
                ret = False
                logger.warning("upgradeutil: remote command '%s' returns unexpected output: %s", cmd, results, exc_info=e)
    return ret


def _get_minimal_seq_in_cluster(nodes) -> typing.Tuple[int, int]:
    try:
        return min(
            _parse_upgrade_seq(stdout.strip()) if rc == 0 else (0, 0)
            for rc, stdout, stderr in _parallax_run(nodes, 'cat {}'.format(SEQ_FILE_PATH)).values()
        )
    except ValueError:
        return 0, 0


def _upgrade(nodes, seq):
    def ask(msg: str):
        pass
    try:
        for key in VERSION_FEATURES.keys():
            if seq < key <= CURRENT_UPGRADE_SEQ:
                for feature_class in VERSION_FEATURES[key]:
                    feature = feature_class()
                    if crmsh.healthcheck.feature_full_check(feature, nodes):
                        logger.debug("upgradeutil: feature '%s' is already functional.", str(feature))
                    else:
                        logger.debug("upgradeutil: fixing feature '%s'...", str(feature))
                        crmsh.healthcheck.feature_fix(feature, nodes, ask)
        logger.debug("upgradeutil: configuration fix succeeded.")
    except crmsh.healthcheck.AskDeniedByUser:
        raise _SkipUpgrade() from None


def upgrade_if_needed():
    if crmsh.config.core.no_ssh:
        return
    if os.geteuid() != 0:
        return
    if not crmsh.utils.can_ask(background_wait=False):
        return
    nodes = crmsh.utils.list_cluster_nodes(no_reg=True)
    if nodes is not None and len(nodes) > 1 \
            and _is_upgrade_needed(nodes):
        logger.debug("upgradeutil: configuration fix needed")
        try:
            if not _is_cluster_target_seq_consistent(nodes):
                logger.warning("crmsh configuration is inconsistent in cluster.")
                raise _SkipUpgrade()
            seq = _get_minimal_seq_in_cluster(nodes)
            logger.debug(
                "Upgrading crmsh from seq %s to %s.",
                seq, _format_upgrade_seq(CURRENT_UPGRADE_SEQ),
            )
            _upgrade(nodes, seq)
        except _SkipUpgrade:
            logger.debug("upgradeutil: configuration fix skipped")
            return
        crmsh.parallax.parallax_call(
            nodes,
            "mkdir -p '{}' && echo '{}' > '{}'".format(
                DATA_DIR,
                _format_upgrade_seq(CURRENT_UPGRADE_SEQ),
                SEQ_FILE_PATH,
            ),
        )
        crmsh.parallax.parallax_call(nodes, 'rm -f {}'.format(FORCE_UPGRADE_FILE_PATH))
        logger.debug("configuration fix finished")


def force_set_local_upgrade_seq():
    """Create the upgrade sequence file and set it to CURRENT_UPGRADE_SEQ.

    It should only be used when initializing new cluster nodes."""
    if not os.path.exists(DATA_DIR):
        crmsh.utils.mkdirs_owned(DATA_DIR, mode=0o755, uid='root', gid='root')
    up_seq = _format_upgrade_seq(CURRENT_UPGRADE_SEQ)
    crmsh.utils.str2file(up_seq, SEQ_FILE_PATH)


def main():
    if sys.argv[1] == 'get-seq':
        print(_format_upgrade_seq(CURRENT_UPGRADE_SEQ))


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
import io
import os
import pwd
import socket
import struct
import subprocess


MSG_EOF  = 0
MSG_USER = 1
MSG_CMD  = 2
MSG_OUT  = 4 
MSG_ERR  = 5
MSG_RC   = 6


class Message:
    @staticmethod
    def write(output, type: int, data: bytes):
        output.write(struct.pack('!ii', type, len(data)))
        output.write(data)

    @staticmethod
    def read(input):
        buf = input.read(8)
        type, length = struct.unpack('!ii', buf)
        if length > 0:
            buf = input.read(length)
        else:
            buf = b''
        return type, buf


class SocketIO(io.RawIOBase):
    def __init__(self, s: socket.socket):
        self._socket = s

    def readable(self) -> bool:
        return True

    def writable(self) -> bool:
        return True

    def read(self, __size: int = -1) -> bytes:
        return self._socket.recv(__size)

    def readinto(self, __buffer) -> int:
        return self._socket.recv_into(__buffer)

    def readall(self) -> bytes:
        raise NotImplementedError

    def write(self, __b) -> int:
        return self._socket.send(__b)


def call(host: str, port: int, cmdline: str):
    family, type, proto, _, sockaddr =  socket.getaddrinfo(host, port, type=socket.SOCK_STREAM)[0]
    with socket.socket(family, type, proto) as s:
        s.connect(sockaddr)
        sout = io.BufferedWriter(SocketIO(s), 4096)
        Message.write(sout, MSG_USER, _getuser().encode('utf-8'))
        Message.write(sout, MSG_CMD, cmdline.encode('utf-8'))
        Message.write(sout, MSG_EOF, b'')
        sout.flush()
        s.shutdown(socket.SHUT_WR)
        rc = None
        stdout = []
        stderr = []
        sin = io.BufferedReader(SocketIO(s), 4096)
        while True:
            type, buf = Message.read(sin)
            if type == MSG_OUT:
                stdout.append(buf)
            elif type == MSG_ERR:
                stderr.append(buf)
            elif type == MSG_RC:
                rc, = struct.unpack('!i', buf)
            elif type == MSG_EOF:
                s.shutdown(socket.SHUT_RD)
                assert rc is not None
                return rc, b''.join(stdout), b''.join(stderr)
            else:
                raise ValueError(f"Unknown message type: {type}")


def serve(stdin, stdout, stderr):
    assert os.geteuid() == 0
    user = None
    cmd = None
    sin = io.BufferedReader(stdin)
    while True:
        type, buf = Message.read(sin)
        if type == MSG_USER:
            user = buf.decode('utf-8')
        elif type == MSG_CMD:
            cmd = buf.decode('utf-8')
        elif type == MSG_EOF:
            assert user is not None
            assert cmd is not None
            break
        else:
            raise ValueError(f"Unknown message type: {type}")
    if user == 'root':
        args = ['/bin/sh']
    else:
        args = ['/bin/su', '-', user, '-c', '/bin/sh']
    result = subprocess.run(
        args,
        input=cmd.encode('utf-8'),
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    sout = io.BufferedWriter(stdout)
    Message.write(sout, MSG_RC, struct.pack('!i', result.returncode))
    Message.write(sout, MSG_OUT, result.stdout)
    Message.write(sout, MSG_ERR, result.stderr)
    Message.write(sout, MSG_EOF, b'')
    stdout.flush()


def _getuser():
    return pwd.getpwuid(os.geteuid()).pw_name


if __name__ == '__main__':
    with open(0, 'rb') as stdin, \
         open(1, 'wb') as stdout, \
         open(2, 'wb') as stderr:
        serve(stdin, stdout, stderr)
